"""Function to scan for pseudolandmarks along the x-axis."""
import cv2
import os
import numpy as np
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _cv2_findcontours, _object_composition
from plantcv.plantcv import params, outputs


def x_axis_pseudolandmarks(img, mask, label=None):
    """Divide up object contour into 20 equidistance segments and generate landmarks for each.

    Parameters
    ----------
    img : numpy.ndarray
        RGB or grayscale image data for plotting
    mask : numpy.ndarray
        Binary mask of the object of interest
    label : str or None, optional
        Optional label parameter, modifies the variable name of
        observations recorded, (default = pcv.params.sample_label)

    Returns
    -------
    top : list
        Top landmark points (depending on axis)
    bottom : list
        Bottom landmark points (depending on axis)
    center_v : list
        Vertical landmark points in middle portion (depending on axis)
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label
    # Find contours
    cnt, cnt_str = _cv2_findcontours(bin_img=mask)
    # Consolidate contours
    obj = _object_composition(contours=cnt, hierarchy=cnt_str)
    # Empty pot type scenario
    if not np.any(obj):
        return ('NA', 'NA'), ('NA', 'NA'), ('NA', 'NA')
    # Bounding rectangle
    _, _, width, _ = cv2.boundingRect(obj)
    extent = width
    # If width is greater than 21 pixels make 20 increments (5% intervals)
    if extent >= 21:
        top, bottom, center_v = _pseudolandmarks(img, obj, mask, label, 0)
        # If the width of the object is less than 20 pixels just make the object a 20 pixel rectangle
    elif extent < 21:
        top, bottom, center_v = _small_img_pseudolandmarks(img, obj, mask, label, 0)

    return top, bottom, center_v


def _small_img_pseudolandmarks(img, obj, mask, label, axis=0):
    """Divide object contour into equidistance segments for small (<=20px tall) images.

    Parameters
    ----------
    img : numpy.ndarray
        RGB or grayscale image data for plotting
    obj : list
        Contours of the object of interest
    mask : numpy.ndarray
        Binary mask of the object of interest
    label : str
        Label parameter, modifies the variable name of
        observations recorded
    axis : int, optional
        Axis along which to generate landmarks, 0 for x axis, 1 for y axis, by default 0

    Returns
    -------
    left_or_top : list
        Left or top landmark points (depending on axis)
    right_or_bottom : list
        Right or bottom landmark points (depending on axis)
    center_h_or_v : list
        Horizontal or vertical landmark points in middle portion (depending on axis)
    """
    # Lists to return
    left_or_top = []
    right_or_bottom = []
    center_h_or_v = []
    # Lists for outputs class
    left_or_top_list = []
    right_or_bottom_list = []
    center_h_or_v_list = []
    # helper lists for x/y directions
    axes_names = ["x", "y"]
    axis_name = axes_names.pop(axis)
    direction_names = [["top", "bottom", "vertical"], ["left", "right", "horizontal"]][axis]
    lt = direction_names[0]
    rb = direction_names[1]
    hv = direction_names[2]
    # If the length of the object is less than 20 pixels make 20 pixel rectangle
    x, y, width, _ = cv2.boundingRect(obj)
    xy = [x, y]
    main_ax = xy.pop(axis)
    other_ax = xy[0]
    coords = list(range(main_ax, main_ax + 20))
    points = [other_ax] * 20
    points2 = [other_ax + width] * 20
    left_or_top = list(zip(points, coords))
    right_or_bottom = list(zip(points2, coords))
    # if using x axis then flip order of points and coords
    if axis:
        left_or_top = list(zip(coords, points))
        right_or_bottom = list(zip(coords, points2))
    left_or_top = np.array(left_or_top)
    left_or_top.shape = (20, 1, 2)
    right_or_bottom = np.array(right_or_bottom)
    right_or_bottom.shape = (20, 1, 2)
    # get moments from mask
    m = cv2.moments(mask, binaryImage=True)
    # Centroid (center of mass)
    cms = (m['m10'] / m['m00'], m['m01'] / m['m00'])
    cm = cms[axis]
    c_points = [cm] * 20
    center_h_or_v = list(zip(c_points, coords))
    center_h_or_v = np.array(center_h_or_v)
    center_h_or_v.shape = (20, 1, 2)
    img2 = np.copy(img)
    for i in left_or_top:
        x = i[0, 0]
        y = i[0, 1]
        cv2.circle(img2, (int(x), int(y)), params.line_thickness, (255, 0, 0), -1)
    for i in right_or_bottom:
        x = i[0, 0]
        y = i[0, 1]
        cv2.circle(img2, (int(x), int(y)), params.line_thickness, (255, 0, 255), -1)
    for i in center_h_or_v:
        x = i[0, 0]
        y = i[0, 1]
        cv2.circle(img2, (int(x), int(y)), params.line_thickness, (0, 79, 255), -1)
    _debug(visual=img2,
           filename=os.path.join(params.debug_outdir, (str(params.device) + f'_{axis_name}_axis_pseudolandmarks.png')))
    # add points to lists for saving output
    for pt in left_or_top:
        left_or_top_list.append(pt[0].tolist())
    for pt in right_or_bottom:
        right_or_bottom_list.append(pt[0].tolist())
    for pt in center_h_or_v:
        center_h_or_v_list.append(pt[0].tolist())
    # save outputs
    outputs.add_observation(sample=label, variable=f'{lt}_lmk', trait=f'{lt} landmark coordinates',
                            method=f'plantcv.plantcv.homology.{axis_name}_axis_pseudolandmarks',
                            scale='none', datatype=tuple,
                            value=tuple(left_or_top_list), label='none')
    outputs.add_observation(sample=label, variable=f'{rb}_lmk', trait=f'{rb} landmark coordinates',
                            method=f'plantcv.plantcv.homology.{axis_name}_axis_pseudolandmarks',
                            scale='none', datatype=tuple,
                            value=tuple(right_or_bottom_list), label='none')
    outputs.add_observation(sample=label, variable=f'center_{hv[0]}_lmk',
                            trait=f'center {hv} landmark coordinates',
                            method=f'plantcv.plantcv.homology.{axis_name}_axis_pseudolandmarks',
                            scale='none', datatype=tuple,
                            value=tuple(center_h_or_v_list), label='none')

    return left_or_top, right_or_bottom, center_h_or_v


def _pseudolandmarks(img, obj, mask, label, axis=0):
    """Divide object contour into equidistance segments for >20px tall images.

    Parameters
    ----------
    img : numpy.ndarray
        RGB or grayscale image data for plotting
    obj : list
        Contours of the object of interest
    mask : numpy.ndarray
        Binary mask of the object of interest
    label : str
        Label parameter, modifies the variable name of
        observations recorded
    axis : int, optional
        Axis along which to generate landmarks, 0 for x axis, 1 for y axis, by default 0

    Returns
    -------
    left_or_top : list
        Left or top landmark points (depending on axis)
    right_or_bottom : list
        Right or bottom landmark points (depending on axis)
    center_h_or_v : list
        Horizontal or vertical landmark points in middle portion (depending on axis)
    """
    x, y, width, height = cv2.boundingRect(obj)
    xy = [x, y]
    main_ax = xy.pop(axis)
    directions = [width, height]
    main_direction = directions.pop(axis)
    channels = [0, 1]
    main_c = channels.pop(axis)
    other_c = channels[0]
    extent = main_direction
    # Lists to return
    left_or_top = []
    right_or_bottom = []
    center_h_or_v = []
    # Lists for outputs class
    left_or_top_list = []
    right_or_bottom_list = []
    center_h_or_v_list = []
    # helper lists for x/y directions
    axes_names = ["x", "y"]
    axis_name = axes_names.pop(axis)
    direction_names = [["top", "bottom", "vertical"], ["left", "right", "horizontal"]][axis]
    lt = direction_names[0]
    rb = direction_names[1]
    hv = direction_names[2]
    # define increment
    inc = int(extent / 21)
    # Define variable for max points and min points
    pts_max = [main_ax]
    pts_min = [main_ax + inc]
    # Get max and min points for each of the intervals
    for i in range(2, 21):
        pt_max = main_ax + (inc * (i - 1))
        pt_min = main_ax + (inc * i)
        # Put these in an array
        pts_max.append(pt_max)
        pts_min.append(pt_min)
    # Combine max and min into a set of tuples
    point_range = list(zip(pts_max, pts_min))
    # define some list variables to fill
    row_median = []
    row_ave = []
    max_width = []
    left_or_top_points = []
    right_or_bottom_points = []
    y_vals = []
    x_centroids = []
    y_centroids = []
    # For each of the 20 intervals
    for pt in point_range:
        # Get the lower and upper bounds
        low_point, high_point = pt
        # Get all rows within these two points
        rows = []
        lps = []
        rps = []
        # Get a continuous list of the values between the top and the bottom of the interval save as vals
        vals = list(range(low_point, high_point))
        # For each row... get all coordinates from object contour that match row
        for v in vals:
            # Value is all entries that match the row
            value = obj[v == obj[:, 0, main_c]]
            if len(value) > 0:
                # Could potentially be more than two points in all contour in each pixel row
                # Grab largest x coordinate (column)
                largest = value[:, 0, other_c].max()
                # Grab smallest x coordinate (column)
                smallest = value[:, 0, other_c].min()
                # Take the difference between the two (this is how far across the object is on this plane)
                row_width = largest - smallest
                # Append this value to a list
                rows.append(row_width)
                lps.append(smallest)
                rps.append(largest)
            if len(value) == 0:
                row_width = 1
                rows.append(row_width)
                lps.append(1)
                rps.append(1)
        # For each of the points find the median and average width
        row_median.append(np.median(np.array(rows)))
        row_ave.append(np.mean(np.array(rows)))
        max_width.append(np.max(np.array(rows)))
        left_or_top_points.append(np.mean(smallest))
        right_or_bottom_points.append(np.mean(largest))
        yval = int((high_point + low_point) / 2)
        y_vals.append(yval)
        # Make a copy of the mask; we want to get landmark points from this
        window = np.copy(mask)
        window[:low_point] = 0
        window[high_point:] = 0
        s = cv2.moments(window)
        # Centroid (center of mass x, center of mass y)
        smx = (largest + smallest) / 2
        smy = yval
        # if difference is very small then recalculate centroids
        if largest - smallest > 3:
            smx, smy = (s['m10'] / max(s['m00'], 0.001),
                        s['m01'] / max(s['m00'], 0.001))
        # append results to centroids
        x_centroids.append(int(smx))
        y_centroids.append(int(smy))
    left_or_top = list(zip(left_or_top_points, y_vals))
    right_or_bottom = list(zip(right_or_bottom_points, y_vals))
    if not axis:
        left_or_top = list(zip(y_vals, left_or_top_points))
        right_or_bottom = list(zip(y_vals, right_or_bottom_points))
    left_or_top = np.array(left_or_top)
    left_or_top.shape = (20, 1, 2)
    right_or_bottom = np.array(right_or_bottom)
    right_or_bottom.shape = (20, 1, 2)
    center_h_or_v = list(zip(x_centroids, y_centroids))
    center_h_or_v = np.array(center_h_or_v)
    center_h_or_v.shape = (20, 1, 2)

    img2 = np.copy(img)
    for i in left_or_top:
        x = i[0, 0]
        y = i[0, 1]
        cv2.circle(img2, (int(x), int(y)), params.line_thickness, (255, 0, 0), -1)
    for i in right_or_bottom:
        x = i[0, 0]
        y = i[0, 1]
        cv2.circle(img2, (int(x), int(y)), params.line_thickness, (255, 0, 255), -1)
    for i in center_h_or_v:
        x = i[0, 0]
        y = i[0, 1]
        cv2.circle(img2, (int(x), int(y)), params.line_thickness, (0, 79, 255), -1)

    _debug(visual=img2,
           filename=os.path.join(params.debug_outdir, (str(params.device) + f'_{axis_name}_axis_pseudolandmarks.png')))

    # Store into global measurements
    for pt in left_or_top:
        left_or_top_list.append(pt[0].tolist())
    for pt in right_or_bottom:
        right_or_bottom_list.append(pt[0].tolist())
    for pt in center_h_or_v:
        center_h_or_v_list.append(pt[0].tolist())

    # save outputs
    outputs.add_observation(sample=label, variable=f'{lt}_lmk', trait=f'{lt} landmark coordinates',
                            method=f'plantcv.plantcv.homology.{axis_name}_axis_pseudolandmarks',
                            scale='none', datatype=tuple,
                            value=tuple(left_or_top_list), label='none')
    outputs.add_observation(sample=label, variable=f'{rb}_lmk', trait=f'{rb} landmark coordinates',
                            method=f'plantcv.plantcv.homology.{axis_name}_axis_pseudolandmarks',
                            scale='none', datatype=tuple,
                            value=tuple(right_or_bottom_list), label='none')
    outputs.add_observation(sample=label, variable=f'center_{hv[0]}_lmk',
                            trait=f'center {hv} landmark coordinates',
                            method=f'plantcv.plantcv.homology.{axis_name}_axis_pseudolandmarks',
                            scale='none', datatype=tuple,
                            value=tuple(center_h_or_v_list), label='none')

    return left_or_top, right_or_bottom, center_h_or_v



# Identify landmark positions within a contour for morphometric analysis

import numpy as np
import math
import cv2
from plantcv.plantcv import params, outputs
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _cv2_findcontours, _object_composition


def acute(img, mask, win, threshold, label=None):
    """
    Identify landmark positions within a contour for morphometric analysis.

    Parameters
    ----------
    img : numpy.ndarray
        Original image used for plotting purposes
    mask : numpy.ndarray
        Binary mask used to generate contour array (necessary for ptvals)
    win : int
        Maximum cumulative pixel distance window for calculating angle score; 1 cm in pixels often works well
    threshold : int
        Angle score threshold to be applied for mapping out landmark coordinate clusters within each contour
    label : str or None, optional
        Optional label parameter, modifies the variable name of
        observations recorded, (default = pcv.params.sample_label)

    Returns
    -------
    homolog_pts : list
        List of pseudo-landmark coordinates (x,y) for each landmark cluster
    start_pts : list
        List of pseudo-landmark island starting coordinates (x,y)
    stop_pts : list
        List of pseudo-landmark island ending coordinates (x,y)
    ptvals : list
        List of mean pixel values underlying each landmark cluster
    chain : list
        List of angle scores for each contour point
    max_dist : list
        List of maximum distances for each contour point
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label
    # initialize returns
    homolog_pts = []
    start_pts = []
    stop_pts = []
    ptvals = []
    max_dist = []
    num_acute_pts = 0
    # Find contours
    contours, hierarchy = _cv2_findcontours(bin_img=mask)
    obj = _object_composition(contours=contours, hierarchy=hierarchy)
    # make list of angles in obj limited by window
    chain = _angle_chain(obj, win)
    index = []
    for c, link in enumerate(chain):
        if float(link) <= threshold:
            index.append(c)

    if len(index) != 0:
        # find islands from acute angles
        isle = _find_islands(obj, index, chain, win)
        # get relevant points from acute islands
        homolog_pts, start_pts, stop_pts, ptvals = _process_islands(obj, isle, chain, mask)
        num_acute_pts = len(homolog_pts)
        ori_img = np.copy(img)
        # Convert grayscale images to color
        if len(np.shape(ori_img)) == 2:
            ori_img = cv2.cvtColor(ori_img, cv2.COLOR_GRAY2BGR)
        # Draw acute points on the original image
        cv2.drawContours(ori_img, homolog_pts, -1, (255, 255, 255), params.line_thickness)
        # print/plot debug image
        _debug(visual=ori_img, filename=f"{params.device}_acute_plms.png")
    # add to outputs and return values for troubleshooting or downstream steps
    outputs.add_observation(sample=label, variable='num_acute_pts', trait='number of acute points',
                            method='plantcv.plantcv.homology.acute', scale='none', datatype=int,
                            value=num_acute_pts, label='none')
    return homolog_pts, start_pts, stop_pts, ptvals, chain, max_dist


def _angle_chain(obj, win):
    """
    Creates a list of angles to be used in detecting acute angles.

    Parameters
    ----------
    obj : plantcv.Objects
        Contours and hierarchy of objects
    win : int
        Maximum cumulative pixel distance window for calculating angle score

    Returns
    -------
    chain : list
        List of angles for each contour point
    """
    # initialize chain list
    chain = []
    # assign 3 points, coordinate by coordinate
    for k in list(range(len(obj))):
        # first point, vert
        vert = obj[k]
        # initial distance is 0, there is only 1 point
        dist_1 = 0
        # loop over points to use with vert, starting by looking backwards for second point (A)
        for r in range(len(obj)):
            rev = k - r
            # get next point of the 3
            pos = obj[rev]
            dist_2 = np.sqrt(np.square(pos[0][0] - vert[0][0]) + np.square(pos[0][1] - vert[0][1]))
            if r >= 2:
                if (dist_2 > dist_1) & (dist_2 <= win):  # Further from vertex than current pt A while within window
                    dist_1 = dist_2
                    pt_a = pos
                elif dist_2 > win:
                    break
            else:
                pt_a = pos
        dist_1 = 0
        # find third point (B), looking forward
        for f in range(len(obj)):
            fwd = k + f
            if fwd >= len(obj):
                fwd -= len(obj)
            pos = obj[fwd]
            dist_2 = np.sqrt(np.square(pos[0][0] - vert[0][0]) + np.square(pos[0][1] - vert[0][1]))
            if f >= 2:
                if (dist_2 > dist_1) & (dist_2 <= win):
                    dist_1 = dist_2
                    pt_b = pos
                elif dist_2 > win:
                    break
            else:
                pt_b = pos

        # Angle in radians derived from Law of Cosines, converted to degrees
        p12 = np.sqrt((vert[0][0]-pt_a[0][0])*(vert[0][0]-pt_a[0][0])+(vert[0][1]-pt_a[0][1])*(vert[0][1]-pt_a[0][1]))
        p13 = np.sqrt((vert[0][0]-pt_b[0][0])*(vert[0][0]-pt_b[0][0])+(vert[0][1]-pt_b[0][1])*(vert[0][1]-pt_b[0][1]))
        p23 = np.sqrt((pt_a[0][0]-pt_b[0][0])*(pt_a[0][0]-pt_b[0][0])+(pt_a[0][1]-pt_b[0][1])*(pt_a[0][1]-pt_b[0][1]))
        dot = (p12*p12 + p13*p13 - p23*p23)/(2*p12*p13)

        ang = math.degrees(math.acos(dot))
        chain.append(ang)
    return chain


def _find_islands(obj, index, chain, win):
    """
    Find islands of acute angles from list of angles.

    Parameters
    ----------
    obj : plantcv.Objects
        Contours and hierarchy of binary mask.
    index : list
        Angles found from contours of mask.
    chain : list
        Angle scores for each contour.
    win : int
        Maximum cumulative pixel distance window for calculating angle score.

    Returns
    -------
    isle : list
        List of islands around acute angles.
    """
    isle = []
    island = []
    # for every acute angle
    for ind in index:
        if not island:
            island.append(ind)
        # if (last element of island) + 1 is this entry, append the entry as is
        elif island[-1] + 1 == ind:
            island.append(ind)
        else:
            pt_a = obj[ind]
            pt_b = obj[island[-1] + 1]
            dist = np.sqrt(np.square(pt_a[0][0] - pt_b[0][0]) + np.square(pt_a[0][1] - pt_b[0][1]))
            if win / 2 > dist:
                island.append(ind)
            else:
                isle.append(island)
                island = [ind]

    isle.append(island)

    if (len(isle) > 1) & (isle[0][0] == 0) & (isle[-1][-1] == (len(chain)-1)):
        island = isle[-1] + isle[0]  # Fuse overlapping ends of contour
        # Delete islands to be spliced if start-end fusion required
        del isle[0]
        del isle[-1]
        isle.insert(0, island)

    return isle


def _process_islands(obj, isle, chain, mask):
    """
    Process list of islands into start/landmark/termination sites.

    Parameters
    ----------
    obj : plantcv.Objects
        Contours and hierarchy of binary mask.
    isle : list
        List of islands around acute angles.
    chain : list
        Angle scores for each contour.
    mask : numpy.ndarray
        Binary mask used to generate contour array.

    Returns
    -------
    homolog_pts : list
        Pseudo-landmarks from each landmark cluster.
    start_pts : list
        Pseudo-landmark island starting position.
    stop_pts : list
        Pseudo-landmark island end position.
    pt_vals  : list
        Point values
    """
    # Initialize empty lists for homologous point max distance method
    maxpts = []
    ss_pts = []
    ts_pts = []
    ptvals = []
    max_dist = [['cont_pos', 'max_dist', 'angle']]
    vals = []
    for island in isle:
        # Identify if contour is concavity/convexity using image mask
        pix_x, pix_y, w, h = cv2.boundingRect(obj[island])  # Obtain local window around island
        for c in range(w):
            for r in range(h):
                # Identify pixels in local window internal to the island hull
                pos = cv2.pointPolygonTest(obj[island], (pix_x + c, pix_y + r), 0)
                if pos > 0:
                    vals.append(mask[pix_y + r][pix_x + c])  # Store pixel value if internal
        if len(vals) > 0:
            ptvals.append(sum(vals) / len(vals))
            vals = []
        else:
            ptvals.append('NaN')
            vals = []
        if len(island) >= 3:               # If landmark is multiple points (distance scan for position)
            ss = obj[island[0]]            # Store isle "x" start site
            ts = obj[island[-1]]           # Store isle "x" termination site
            dist_1 = 0
            for d in island:   # Scan from ss to ts within isle "x"
                site = obj[[d]]
                ss_d = np.sqrt(np.square(ss[0][0] - site[0][0][0]) + np.square(ss[0][1] - site[0][0][1]))
                ts_d = np.sqrt(np.square(ts[0][0] - site[0][0][0]) + np.square(ts[0][1] - site[0][0][1]))
                # Current mean distance of 'd' to 'ss' & 'ts'
                dist_2 = np.mean([np.abs(ss_d), np.abs(ts_d)])
                max_dist.append([d, dist_2, chain[d]])
                if dist_2 > dist_1:                          # Current mean distance better fit that previous best?
                    pt = d
                    dist_1 = dist_2                          # Current mean becomes new best mean
            # print pt
            if params.verbose:
                print(f"Landmark site: {pt}, Start site: {island[0]}, Term. site: {island[-1]}")

            maxpts.append(pt)           # Empty 'pts' prior to next mean distance scan
            ss_pts.append(island[0])
            ts_pts.append(island[-1])

        if params.verbose:
            print(f'Landmark point indices: {maxpts}')
            print(f'Starting site indices: {ss_pts}')
            print(f'Termination site indices: {ts_pts}')
    homolog_pts = obj[maxpts]
    start_pts = obj[ss_pts]
    stop_pts = obj[ts_pts]

    return homolog_pts, start_pts, stop_pts, ptvals



"""Help visualize histograms for hyperspectral images."""
import os
import numpy as np
import pandas as pd
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import fatal_error, params, color_palette
from plantcv.plantcv.hyperspectral import _find_closest
from plantcv.plantcv.visualize import histogram
import altair as alt
import math


def _get_color_dict_uv():
    """Create a color dictionary for UV wavelengths."""
    params.color_scale = "cool_r"
    uv_wavelengths = np.arange(290, 445)
    uv_colors_ = color_palette(num=256)
    uv_colors_ = uv_colors_[0:len(uv_wavelengths)]
    uv_colors_ = [tuple(xi / 255 for xi in x) for x in uv_colors_[::-1]]
    uv_colors = {}
    for i, wv in enumerate(uv_wavelengths):
        uv_colors[wv] = uv_colors_[i]
    return uv_colors


def _get_color_dict_vis():
    """Create a color dictionary for VIS wavelengths."""
    params.color_scale = "turbo"
    vis_wavelengths = np.arange(445, 701)
    vis_colors_ = color_palette(num=256)
    vis_colors_ = [tuple(xi / 255 for xi in x) for x in vis_colors_]
    vis_colors = {}
    for i, wv in enumerate(vis_wavelengths):
        vis_colors[wv] = vis_colors_[i]
    return vis_colors


def _get_color_dict_nir():
    """Create a color dictionary for infrared wavelengths."""
    params.color_scale = "inferno"
    nir_wavelengths = np.arange(701, 1725)
    nir_colors_ = color_palette(num=256)
    nir_colors_ = [tuple(xi / 255 for xi in nir_colors_[math.floor(idx / 4)]) for (idx, _) in
                   enumerate(nir_wavelengths)]
    nir_colors = {}
    for i, wv in enumerate(nir_wavelengths):
        nir_colors[wv] = nir_colors_[i]
    return nir_colors


def _rgb_to_webcode(rgb_values):
    """
    Convert RGB values to webcodes.

    Inputs:
    rgb_value: a tuple of RGB values (0~1, float)

    Returns:
    webcode: a webcode string encoding the RGB values
    """
    webcode = "#"
    for value in rgb_values:
        code_ = hex(int(value*255)).replace('0x', '')
        code = code_.upper() if len(code_) > 1 else f'0{code_.upper()}'
        webcode += code
    return webcode


# Helper function to update color dictionary
def update_color_dict(wavelengths, color):
    """Update the color dictionary with the given color."""
    return {wv: color for wv in wavelengths}


def _get_color_dict(match_wls):
    """
    Sets the color scale based on the wavelength range.

    Inputs:
    match_wls: list of wavelengths to be plotted

    Returns:
    color_dict: dictionary of wavelengths and their corresponding colors
    """
    color_dict = {}

    color_dict.update(_handle_under_uv(match_wls))
    color_dict.update(_handle_uv(match_wls))
    color_dict.update(_handle_visible(match_wls))
    color_dict.update(_handle_nir(match_wls))
    color_dict.update(_handle_above_nir(match_wls))

    return color_dict


def _handle_under_uv(match_wls):
    """Helper to handle under wavelengths under uv"""
    under_uv_wls = [x for x in match_wls if x < 290]
    if under_uv_wls:
        params.color_scale = "cool_r"
        color_ = color_palette(num=256)[-154]
        return update_color_dict(under_uv_wls, color_)
    return {}


def _handle_uv(match_wls):
    """Helper to handle uv wavelengths"""
    uv_wls = [x for x in match_wls if 290 <= x < 445]
    if uv_wls:
        return _get_color_dict_uv()
    return {}


def _handle_visible(match_wls):
    """Helper to handle visible wavelengths"""
    vis_wls = [x for x in match_wls if 445 <= x < 701]
    if vis_wls:
        return _get_color_dict_vis()
    return {}


def _handle_nir(match_wls):
    """Helper to handle nir wavelengths"""
    nir_wls = [x for x in match_wls if 701 <= x < 1701]
    if nir_wls:
        return _get_color_dict_nir()
    return {}


def _handle_above_nir(match_wls):
    """Helper to handle wavelengths above nir"""
    above_nir_wls = [x for x in match_wls if x >= 1701]
    if above_nir_wls:
        params.color_scale = "inferno"
        color_ = color_palette(num=256)[-1]
        return update_color_dict(above_nir_wls, color_)
    return {}


def hyper_histogram(hsi, mask=None, bins=100, lower_bound=None, upper_bound=None,
                    title=None, wvlengths=None):
    """Plot a histograms of selected wavelengths from a hyperspectral image.

    This function calculates the histogram of selected wavelengths hyperspectral images
    The color of the histograms are based on the wavelength if the wavelength is in the range of visible spectrum;
    otherwise, random colors are assigned

    Inputs:
    hsi          = Hyperspectral data instance
    mask         = binary mask, if provided, calculate histogram from masked area only (default=None)
    bins         = divide the data into n evenly spaced bins (default=100)
    lower_bound  = the lower bound of the bins (x-axis min value) (default=None)
    upper_bound  = the upper bound of the bins (x-axis max value) (default=None)
    title        = a custom title for the plot (default=None)
    wvlengths    = (optional) list of wavelengths to show histograms (default = [480,550,670],  i.e. only show
                                                                    histograms for blue, green, and red bands)

    Returns:
    fig_hist      = histogram figure

    :param hsi: plantcv.plantcv.classes.Spectral_data
    :param mask: numpy.ndarray
    :param bins: int
    :param lower_bound: None, int, float
    :param upper_bound: None, int, float
    :param title: None, str
    :param wvlengths: list
    :return fig_hist: altair.vegalite.v5.api.Chart
    """
    # Set default wavelengths to plot
    if wvlengths is None:
        wvlengths = [480, 550, 650]

    # Always sort desired wavelengths
    wvlengths.sort()

    # Available wavelengths of the spectral data
    wl_keys = hsi.wavelength_dict.keys()
    wls = np.array([float(i) for i in wl_keys])
    wls.sort()

    # Spectral resolution of the spectral data
    diffs = [wls[i] - wls[i - 1] for i in range(1, len(wls))]
    spc_res = sum(diffs) / len(diffs)

    # Check if the distance is greater than 2x the spectral resolution
    # If the distance > 2x resolution, it is considered being out of available ranges
    checks = []
    for i in range(0, len(wvlengths)):
        checks.append(hsi.min_wavelength - wvlengths[i] > 2 * spc_res)
        checks.append(wvlengths[i] - hsi.max_wavelength > 2 * spc_res)
    if np.any(checks):
        fatal_error(f"At least one band is too far from the available wavelength range: "
                    f"({hsi.min_wavelength},{hsi.max_wavelength})!")

    # Find indices of bands whose wavelengths are closest to desired ones
    match_ids = [_find_closest(wls, wv) for wv in wvlengths]
    match_wls = [round(wls[i]) for i in match_ids]

    # prepare color dictionary(ies)
    color_dict = _get_color_dict(match_wls)
    colors_rgb = [color_dict[wv] for wv in match_wls]
    colors_hex = [_rgb_to_webcode(x) for x in colors_rgb]

    array_data = hsi.array_data

    # List of wavelengths recorded created from parsing the header file will be string, make list of floats
    histograms = {}
    hist_dataset = pd.DataFrame(columns=['reflectance'])
    debug = params.debug
    params.debug = None

    # Create a dataframe for all histogram related information
    # (using the "histogram" function in "visualization" subpackage)
    for i_wv, (wv, color) in enumerate(zip(wvlengths, colors_hex)):
        idx = match_ids[i_wv]
        _, hist_data = histogram(array_data[:, :, idx], mask=mask, bins=bins, lower_bound=lower_bound,
                                 upper_bound=upper_bound, title=title, hist_data=True)
        histograms[wv] = {"label": wv, "graph_color": color, "reflectance": hist_data['pixel intensity'].tolist(),
                          "hist": hist_data['proportion of pixels (%)'].tolist()}
        if i_wv == 0:
            hist_dataset['reflectance'] = hist_data['pixel intensity'].tolist()
        hist_dataset[wv] = hist_data['proportion of pixels (%)'].tolist()

    # Make the histogram figure using pandas and altair
    df_hist = pd.melt(hist_dataset, id_vars=['reflectance'], value_vars=wvlengths,
                      var_name='Wavelength (' + hsi.wavelength_units + ')', value_name='proportion of pixels (%)')

    fig_hist = alt.Chart(df_hist).mark_line().encode(
        x="reflectance",
        y="proportion of pixels (%)",
        color=alt.Color('Wavelength (' + hsi.wavelength_units + ')').scale(scheme='turbo'),
        tooltip=['Wavelength (' + hsi.wavelength_units + ')']
        )

    if title is not None:
        fig_hist.properties(title=title)

    params.debug = debug
    _debug(fig_hist, filename=os.path.join(params.debug_outdir, str(params.device) + '_histogram.png'))

    return fig_hist.interactive()



# Correct for nonuniform illumination

import os
import cv2
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv import gaussian_blur
from plantcv.plantcv.transform import rescale
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _rgb2gray


def nonuniform_illumination(img, ksize):
    """Correct for non uniform illumination i.e. spotlight correction.

    Inputs:
    img       = RGB or grayscale image data
    ksize     = (optional) new minimum value for range of interest. default = 0

    Returns:
    corrected_img = rescaled image

    :param img: numpy.ndarray
    :param ksize: int
    :return corrected_img: numpy.ndarray
    """
    if len(np.shape(img)) == 3:
        img = _rgb2gray(img)

    # Fill foreground objects
    kernel = np.ones((ksize, ksize), np.uint8)
    opening = cv2.morphologyEx(img, cv2.MORPH_OPEN, kernel)

    # Store debug mode
    debug = params.debug
    params.debug = None

    # Heavily blurred image acts like a background image
    blurred_img = gaussian_blur(opening, ksize=(ksize, ksize))
    img_mean = np.mean(blurred_img)
    corrected_img = img - blurred_img + img_mean
    corrected_img = rescale(gray_img=corrected_img, min_value=0, max_value=255)

    # Reset debug mode
    params.debug = debug

    _debug(visual=corrected_img, filename=os.path.join(params.debug_outdir, str(params.device) + '_correct_illumination.png'))

    return corrected_img



"""Find angles in degrees of skeleton segments."""
import os
import cv2
import numpy as np
import pandas as pd
from plantcv.plantcv import params
from plantcv.plantcv import outputs
from plantcv.plantcv import color_palette
from plantcv.plantcv._debug import _debug


def segment_angle(segmented_img, objects, label=None):
    """
    Calculate the angle of skeleton segments in degrees by fitting a linear regression line.

    Parameters
    ----------
    segmented_img : numpy.ndarray
        Segmented image to plot slope lines and angles on.
    objects : list
        List of contours representing skeleton segments.
    label : str, optional
        Label for the observation variable. If None, uses `params.sample_label`.

    Returns
    -------
    labeled_img : numpy.ndarray
        Segmented image with slope lines and angles labeled.
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    label_coord_x = []
    label_coord_y = []
    segment_angles = []

    labeled_img = segmented_img.copy()

    # Use a previously saved color scale if available
    rand_color = color_palette(num=len(objects), saved=True)

    for i, cnt in enumerate(objects):
        # Find bounds for regression lines to get drawn
        rect = cv2.minAreaRect(cnt)
        pts = cv2.boxPoints(rect)
        df = pd.DataFrame(pts, columns=('x', 'y'))
        x_max = int(df['x'].max())
        x_min = int(df['x'].min())

        # Find line fit to each segment
        [vx, vy, x, y] = cv2.fitLine(objects[i], cv2.DIST_L2, 0, 0.01, 0.01)
        slope = -vy / vx
        left_list = int(np.array(((x - x_min) * slope) + y).item())
        right_list = int(np.array(((x - x_max) * slope) + y).item())

        if abs(slope) > 1000000:
            print("Slope of contour with ID#", i, "is", slope, "and cannot be plotted.")
        else:
            # Draw slope lines
            cv2.line(labeled_img, (x_max - 1, right_list), (x_min, left_list), rand_color[i], 1)

        # Store coordinates for labels
        label_coord_x.append(objects[i][0][0][0])
        label_coord_y.append(objects[i][0][0][1])

        # Calculate degrees from slopes
        angle_value = (np.arctan(slope[0]) * 180 / np.pi).astype(np.float64)
        segment_angles.append(angle_value)

    segment_ids = []
    for i, cnt in enumerate(objects):
        # Label slope lines
        w = label_coord_x[i]
        h = label_coord_y[i]
        text = f"{segment_angles[i]:0,.2f}"
        cv2.putText(img=labeled_img, text=text, org=(w, h), fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                    fontScale=params.text_size, color=(150, 150, 150), thickness=params.text_thickness)
        segment_ids.append(i)

    outputs.add_observation(sample=label, variable='segment_angle', trait='segment angle',
                            method='plantcv.plantcv.morphology.segment_angle', scale='degrees', datatype=list,
                            value=segment_angles, label=segment_ids)

    _debug(visual=labeled_img, filename=os.path.join(params.debug_outdir, f"{params.device}_segmented_angles.png"))

    return labeled_img



import os
import random
import shutil
import json
from plantcv.plantcv import fatal_error


def sample_images(source_path, dest_path, num=100):
    """Gets a sample of images from the source directory and copies them to the destination directory.

    Parameters
    ----------
    source_path : str
        The directory containing the images to be sampled
    dest_path : str
        The directory where the sampled images will be copied
    num : int, optional
        The number of images to sample, by default 100

    Raises
    ------
    IOError
        Raised if the source directory does not exist
    """
    if not os.path.exists(source_path):
        raise IOError(f"Directory does not exist: {source_path}")

    if not os.path.exists(dest_path):
        os.makedirs(dest_path)  # exist_ok argument does not exist in python 2

    # If SnapshotInfo exists then need to make a new csv for the random image sample
    if os.path.exists(os.path.join(source_path, 'SnapshotInfo.csv')):
        _sample_phenofront(source_path, dest_path, num)
    elif os.path.exists(os.path.join(source_path, 'metadata.json')):
        _sample_phenodata(source_path, dest_path, num)
    else:
        _sample_filenames(source_path, dest_path, num)


def _sample_phenofront(source_path, dest_path, num=100):
    """
    Sample images from a phenofront dataset.
    :param source_path: Path to phenofront images.
    :param dest_path: Path to save sampled images.
    :param num: Number of images to sample.
    :return: None
    """
    line_array = []
    with open(os.path.join(source_path, 'SnapshotInfo.csv')) as fp:
        header = fp.readline()
        for line in fp:
            line = line.rstrip("\n")
            element_arr = line.split(',')
            if element_arr[-1]:
                line_array.append(element_arr)

    # Check to make sure number of imgs to select is less than number of images found
    if num > len(line_array):
        fatal_error(f"Number of snapshots found ({len(line_array)}) less than 'num'.")

    # Create SnapshotInfo file
    with open(os.path.join(dest_path, 'SnapshotInfo.csv'), 'w') as out_file:
        out_file.write(header)

        # Get random snapshots
        random_index = random.sample(range(0, len(line_array) - 1), num)
        for i in random_index:
            row = line_array[int(i)]
            out_file.write(','.join(row) + "\n")
            snap_path = os.path.join(source_path, "snapshot" + row[1])
            folder_path = os.path.join(dest_path, "snapshot" + row[1])
            if not os.path.exists(folder_path):
                os.mkdir(folder_path)  # the beginning of folder_path (dest_path) already exists from above
            for root, _, files in os.walk(snap_path):
                for file in files:
                    shutil.copy(os.path.join(root, file), folder_path)


def _sample_filenames(source_path, dest_path, num=100):
    """
    Sample images from a filenames dataset.
    :param source_path: Path to images.
    :param dest_path: Path to save sampled images.
    :param num: Number of images to sample.
    :return: None
    """
    img_element_array = []
    img_extensions = ['.png', '.jpg', '.jpeg', '.tif', '.tiff', '.gif']
    for root, _, files in os.walk(source_path):
        for file in files:
            # Check file type so that only images get copied over
            ext = os.path.splitext(file)[1]
            if ext.lower() in img_extensions:
                img_element_array.append(os.path.join(root, file))

    # Check to make sure number of imgs to select is less than number of images found
    if num > len(img_element_array):
        fatal_error(f"Number of images found ({len(img_element_array)}) less than 'num'.")

    # Get random images
    random_index = random.sample(range(0, len(img_element_array) - 1), num)
    # Copy images over to destination
    for i in random_index:
        shutil.copy(img_element_array[int(i)], dest_path)


def _sample_phenodata(source_path, dest_path, num=100):
    """
    Sample images from a phenodata dataset.
    :param source_path: Path to phenodata images.
    :param dest_path: Path to save sampled images.
    :param num: Number of images to sample.
    :return: None
    """
    # Initialize an empty dataset
    sampled_dataset = {}
    # Read in the metadata
    with open(os.path.join(source_path, "metadata.json"), "r") as fp:
        dataset = json.load(fp)
    # Set the dataset to the sampled dataset
    sampled_dataset["dataset"] = dataset["dataset"]
    # Leave the environment secton empty
    sampled_dataset["environment"] = {}
    # Initialize the images section
    sampled_dataset["images"] = {}
    # Create a unique dictionary of snapshot IDs
    snapshots = {}
    # Store the snapshot IDs in the snapshots dictionary
    for value in dataset["images"].values():
        snapshots[value["snapshot"]] = True
    # Check to make sure number of imgs to select is less than number of images found
    if len(snapshots) < num:
        fatal_error(f"Number of snapshots found ({len(snapshots)}) less than 'num'.")
    # Randomly select the snapshots
    random_snapshots = random.sample(list(snapshots.keys()), num)
    # Iterate over all images in the dataset
    for fpath, meta in dataset["images"].items():
        # If the snapshot ID is in the random snapshots
        if meta["snapshot"] in random_snapshots:
            # Store the image in the sampled dataset
            sampled_dataset["images"][fpath] = meta
            # Copy the image to the destination directory
            parent_path = os.path.split(fpath)[0]
            os.makedirs(os.path.join(dest_path, parent_path), exist_ok=True)
            shutil.copy(os.path.join(source_path, fpath), os.path.join(dest_path, fpath))
    # Write the sampled dataset to a JSON file
    with open(os.path.join(dest_path, "metadata.json"), "w") as fp:
        json.dump(sampled_dataset, fp, indent=4)



# Generate a plm multivariate space for downstream use in homology group assignments

import numpy as np
from plantcv.plantcv import params


def space(cur_plms, include_bound_dist=False, include_centroid_dist=False, include_orient_angles=False):
    """
    Generate a plm multivariate space for downstream use in homology group assignments

    Inputs:
    cur_plms    = A pandas array of acute plms representing capturing two adjacent frames in a time series
                  or otherwise analogous dataset in order to enable homology assignments
    include_bound_dist = Add bounding box distances to space for Starscape clustering
    include_centroid_dist = Add centroid distances to space for Starscape clustering
    include_orient_angles = Add plm and centroid orientation angles to space for Starscape clustering

    :param cur_plms: pandas.core.frame.DataFrame
    :param include_bound_dist: bool
    :param include_centroid_dist: bool
    :param include_orient_angles: bool
    :return new_plms: pandas.core.frame.DataFrame
    """
    new_plms = cur_plms.copy(deep=True)
    bot_left = [int(min(new_plms.loc[:, ['plm_x']].values).item()), int(max(new_plms.loc[:, ['plm_y']].values).item())]
    bot_right = [int(max(new_plms.loc[:, ['plm_x']].values).item()), int(max(new_plms.loc[:, ['plm_y']].values).item())]
    top_left = [int(min(new_plms.loc[:, ['plm_x']].values).item()), int(min(new_plms.loc[:, ['plm_y']].values).item())]
    top_right = [int(max(new_plms.loc[:, ['plm_x']].values).item()), int(min(new_plms.loc[:, ['plm_y']].values).item())]
    centroid = [int(np.mean(new_plms.loc[:, ['plm_x']].values).item()), int(np.mean(new_plms.loc[:, ['plm_y']].values).item())]

    bot_left_dist = np.sqrt(np.square(new_plms.loc[:, ['plm_x']].values - bot_left[0]) + np.square(
        new_plms.loc[:, ['plm_y']].values - bot_left[1]))
    bot_right_dist = np.sqrt(np.square(new_plms.loc[:, ['plm_x']].values - bot_right[0]) + np.square(
        new_plms.loc[:, ['plm_y']].values - bot_right[1]))
    top_left_dist = np.sqrt(np.square(new_plms.loc[:, ['plm_x']].values - top_left[0]) + np.square(
        new_plms.loc[:, ['plm_y']].values - top_left[1]))
    top_right_dist = np.sqrt(np.square(new_plms.loc[:, ['plm_x']].values - top_right[0]) + np.square(
        new_plms.loc[:, ['plm_y']].values - top_right[1]))

    if include_bound_dist is True:
        new_plms.insert(len(new_plms.columns), 'bot_left_dist', bot_left_dist, True)
        new_plms.insert(len(new_plms.columns), 'bot_right_dist', bot_right_dist, True)
        new_plms.insert(len(new_plms.columns), 'top_left_dist', top_left_dist, True)
        new_plms.insert(len(new_plms.columns), 'top_right_dist', top_right_dist, True)

    centroid_dist = np.sqrt(np.square(new_plms.loc[:, ['plm_x']].values - centroid[0]) + np.square(
        new_plms.loc[:, ['plm_y']].values - centroid[1]))

    if include_centroid_dist is True:
        new_plms.insert(len(new_plms.columns), 'centroid_dist', centroid_dist, True)

    run = (
                  (new_plms.loc[:, ['SS_x']].values + new_plms.loc[:, ['TS_x']].values) / 2
          ) - new_plms.loc[:, ['plm_x']].values
    rise = (
                   (new_plms.loc[:, ['SS_y']].values + new_plms.loc[:, ['TS_y']].values) / 2
           ) - new_plms.loc[:, ['plm_y']].values

    centroid_run = (new_plms.loc[:, ['plm_x']].values - centroid[0])
    centroid_rise = (new_plms.loc[:, ['plm_y']].values - centroid[1])

    orientation = []
    centroid_orientation = []

    for m in range(0, len(run)):
        # Use the sign of the run to determine if the weight should shift
        # in the 0-180 or 180-360 range for 360 arc conversion
        a = 0
        if run[m] > 0:
            a = 90 - (np.arctan(rise[m] / run[m]) * (180 / np.pi))
        elif run[m] < 0:
            a = -(90 - (np.arctan(rise[m] / run[m]) * (180 / np.pi)))
        orientation.append(float(np.array(a).item()))

        # Use the sign of the run to determine if the weight should
        # shift in the 0-180 or 180-360 range for 360 arc conversion
        centroid_a = 0
        if centroid_run[m] > 0:
            centroid_a = 90 - (np.arctan(centroid_rise[m] / centroid_run[m]) * (180 / np.pi))
        elif centroid_run[m] < 0:
            centroid_a = -(90 - (np.arctan(centroid_rise[m] / centroid_run[m]) * (180 / np.pi)))
        centroid_orientation.append(float(np.array(centroid_a).item()))

    if include_orient_angles is True:
        new_plms.insert(len(new_plms.columns), 'orientation', orientation, True)
        new_plms.insert(len(new_plms.columns), 'centroid_orientation', centroid_orientation, True)

    if params.debug is not None:
        print(new_plms.head())

    return new_plms



"""A function that analyzes the width of an object using a distance transformation."""

import os
import cv2
import numpy as np
from plantcv.plantcv import outputs, params
from plantcv.plantcv._helpers import _scale_size, _dilate
from plantcv.plantcv._debug import _debug


def segment_width(segmented_img, skel_img, labeled_mask, n_labels=1, label=None):
    """
    A function that analyzes the width of an object using a distance transformation.

    Parameters
    ----------
    segmented_img : numpy.ndarray
        Segmented image to plot slope lines and angles on.
    skel_img : numpy.ndarray
        Skeletonized image, ideally pruned/clean
    labeled_mask : numpy.ndarray
        Labeled mask of objects (32-bit).
    n_labels : int
        Total number expected individual objects (default = 1).
    label : str, optional
        Label for the observation variable. If None, uses `params.sample_label`.

    Returns
    -------
    labeled_img : numpy.ndarray
        Segmented image with average width per segment.
    """
    widths = []
    stdevs = []
    max_width = []
    dilated_midline = _dilate(skel_img, 2, 1)
    labeled_img = segmented_img.copy()

    for i in range(1, n_labels + 1):
        submask = np.where(labeled_mask == i, 255, 0).astype(np.uint8)
        if np.count_nonzero(submask) > 0:
            mask_copy = submask.copy().astype(np.uint8)
            # Find contours from the submask
            id_objects = cv2.findContours(mask_copy, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)[-2]
            # ID and store area values and centers of mass for labeling them
            m = cv2.moments(id_objects[0])
            # Skip iteration if contour area is zero
            # This is needed because cv2.contourArea can be > 0 while moments area is 0.
            if m['m00'] != 0:
                label_coord_x = int(m["m10"] / m["m00"])
                label_coord_y = int(m["m01"] / m["m00"])
            k = cv2.distanceTransform(mask_copy, cv2.DIST_L2, cv2.DIST_MASK_PRECISE)

            # Select central non-zero values from weighted distance transform
            nonzero_mask = (k * dilated_midline) != 0
            weighted_values = k[nonzero_mask]

            if len(weighted_values) > 0:
                stroke_width = 2 * np.mean(weighted_values)
                stroke_width_max = 2 * np.max(weighted_values)
                width_std = np.std(weighted_values)
                widths.append(stroke_width.astype(np.float64))
                stdevs.append(width_std.astype(np.float64))
                max_width.append(stroke_width_max.astype(np.float64))
                text = str(int(stroke_width))
                if params.verbose:
                    cv2.putText(img=labeled_img, text=text,
                                org=(label_coord_x, label_coord_y),
                                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                                fontScale=params.text_size, color=(150, 150, 150),
                                thickness=params.text_thickness)
            else:
                widths.append(0)
                stdevs.append("NA")
                max_width.append(0)

    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label
    outputs.add_observation(sample=label, variable='mean_segment_width', trait='mean segment widths',
                            method='plantcv.plantcv.morphology.segment_width',
                            scale=params.unit, datatype=list,
                            value=_scale_size(value=widths, trait_type="linear"),
                            label=list(range(1, n_labels + 1)))
    outputs.add_observation(sample=label, variable='segment_width_std', trait='segment width standard deviation',
                            method='plantcv.plantcv.morphology.segment_width',
                            scale="pixels", datatype=list,
                            value=stdevs,
                            label=list(range(1, n_labels + 1)))
    outputs.add_observation(sample=label, variable='segment_width_max', trait='maximum segment widths',
                            method='plantcv.plantcv.morphology.segment_width',
                            scale=params.unit, datatype=list,
                            value=_scale_size(value=max_width, trait_type="linear"),
                            label=list(range(1, n_labels + 1)))
    # Debugging
    _debug(visual=labeled_img, filename=os.path.join(params.debug_outdir, str(params.device) + '_segment_width.png'))

    return labeled_img



"""Find euclidean lengths of skeleton segments."""
import os
import cv2
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv import outputs
from plantcv.plantcv import fatal_error
from plantcv.plantcv import color_palette
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _cv2_findcontours, _find_tips, _scale_size
from scipy.spatial.distance import euclidean


def segment_euclidean_length(segmented_img, objects, label=None):
    """Use segmented skeleton image to gather euclidean length measurements per segment.

    Inputs:
    segmented_img = Segmented image to plot lengths on
    objects       = List of contours
    label         = Optional label parameter, modifies the variable name of
                    observations recorded (default = pcv.params.sample_label).

    Returns:
    labeled_img      = Segmented debugging image with lengths labeled

    :param segmented_img: numpy.ndarray
    :param objects: list
    :param label: str
    :return labeled_img: numpy.ndarray
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    x_list = []
    y_list = []
    segment_lengths = []
    # Create a color scale, use a previously stored scale if available
    rand_color = color_palette(num=len(objects), saved=True)

    labeled_img = segmented_img.copy()
    # Store debug
    debug = params.debug
    params.debug = None

    for i, obj in enumerate(objects):
        # Store coordinates for labels
        x_list.append(obj[0][0][0])
        y_list.append(obj[0][0][1])

        # Draw segments one by one to group segment tips together
        finding_tips_img = np.zeros(segmented_img.shape[:2], np.uint8)
        cv2.drawContours(finding_tips_img, objects, i, (255, 255, 255), 1, lineType=8)
        segment_tips, _, _ = _find_tips(finding_tips_img)
        tip_objects, _ = _cv2_findcontours(bin_img=segment_tips)
        points = []
        if not len(tip_objects) == 2:
            fatal_error("Too many tips found per segment, try pruning again")
        for t in tip_objects:
            # Gather pairs of coordinates
            x, y = t.ravel()
            coord = (x, y)
            points.append(coord)

        # Draw euclidean distance lines
        cv2.line(labeled_img, points[0], points[1], rand_color[i], 1)

        # Calculate euclidean distance between tips of each contour
        segment_lengths.append(float(euclidean(points[0], points[1])))

    segment_ids = []
    # Reset debug mode
    params.debug = debug

    # Put labels of length
    for c, value in enumerate(segment_lengths):
        text = f"{value:0,.2f}"
        w = x_list[c]
        h = y_list[c]
        cv2.putText(img=labeled_img, text=text, org=(w, h), fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                    fontScale=params.text_size, color=(150, 150, 150), thickness=params.text_thickness)
        segment_ids.append(c)

    outputs.add_observation(sample=label, variable='segment_eu_length', trait='segment euclidean length',
                            method='plantcv.plantcv.morphology.segment_euclidean_length', scale=params.unit, datatype=list,
                            value=_scale_size(segment_lengths), label=segment_ids)

    _debug(visual=labeled_img, filename=os.path.join(params.debug_outdir, f"{params.device}_segment_eu_lengths.png"))

    return labeled_img



# Tile output images in a plot to visualize all at once

import cv2
import numpy as np
import os
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug


def _row_resize(row, ncol):
    """Resizes and concatenates objects in a row.

    Parameters
    ----------
    row : list of numpy.ndarray
        List of images to concatenate.
    ncol : int
        Number of columns in desired composite image.

    Returns
    -------
    numpy.ndarray
        Image concatenated horizontally.
    """
    h_min = min(img.shape[0] for img in row)
    # Resizing each image so they're the same
    row_resize = [cv2.resize(img, (int(img.shape[1] * h_min / img.shape[0]), h_min),
                             interpolation=cv2.INTER_CUBIC) for img in row]
    # Add empty images to the end of the row so things still stay the same size
    while len(row_resize) < ncol:
        row_resize.append(np.zeros(row_resize[0].shape, dtype=np.uint8))
    # Concatenate horizontally
    return cv2.hconcat(row_resize)


# Same as _row_resize but for columns
def _col_resize(col):
    """Resized and concatenates objects in a column.

    Parameters
    ----------
    col : list of numpy.ndarray
        List of images to concatenate vertically.

    Returns
    -------
    numpy.ndarray
        Image concatenated vertically.
    """
    w_min = min(img.shape[1] for img in col)
    col_resize = [cv2.resize(img, (w_min, int(img.shape[0] * w_min / img.shape[1])),
                             interpolation=cv2.INTER_CUBIC) for img in col]
    return cv2.vconcat(col_resize)


# The function that does the tiling
def tile(images, ncol):
    """Tile a list of images into a composite with given dimensions.

    Parameters
    ----------
    images : list of numpy.ndarray
        List of images to tile.
    ncol : int
        Number of columns in desired composite image.

    Returns
    -------
    numpy.ndarray
        Tiled composite image.
    """
    # Increment the device counter
    params.device += 1

    # Calculate number of rows - always rounds up
    nrow = int(len(images) / ncol) + (len(images) % ncol > 0)
    tracker = 0
    mat = []
    for _ in range(nrow):
        row = []
        for _ in range(ncol):
            if tracker <= (len(images) - 1):
                row.append(images[tracker])
            tracker += 1
        mat.append(_row_resize(row, ncol))
    comp_img = _col_resize(mat)
    _debug(visual=comp_img, filename=os.path.join(params.debug_outdir, f"{params.device}_tile.png"))
    return comp_img



# Crop position mask

import cv2
import numpy as np
import math
import os
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _cv2_findcontours, _grayscale_to_rgb
from plantcv.plantcv import fatal_error
from plantcv.plantcv import params


def crop_position_mask(img, mask, x, y, v_pos="top", h_pos="right"):
    """
    Crop a binary mask to a specified position within an image.

    Parameters
    ----------
    img : numpy.ndarray
        RGB or grayscale image data for plotting.
    mask : numpy.ndarray
        Binary mask to use (must be correct size; if not, use make_resize_mask function).
    x : int
        x position for cropping.
    y : int
        y position for cropping.
    v_pos : str, optional
        Vertical position to push from, either "top" or "bottom" (default is "top").
    h_pos : str, optional
        Horizontal position to push to, either "right" or "left" (default is "right").

    Returns
    -------
    newmask : numpy.ndarray
        Cropped image mask.

    Raises
    ------
    ValueError
        If x or y are negative or non-integer values.
    ValueError
        If v_pos or h_pos are not valid options.
    ValueError
        If mask is not a binary image.

    Notes
    -----
    - The function centers and resizes the mask to match the image size.
    - The mask is cropped and positioned according to the specified vertical and horizontal parameters.
    - Debug images are saved if debugging is enabled.
    """
    if not all([x > 0, y > 0]):
        fatal_error("x and y cannot be negative numbers or non-integers")

    if v_pos.upper() not in ["TOP", "BOTTOM"]:
        fatal_error(f'{v_pos} is not valid, must be "top" or "bottom"!')

    if h_pos.upper() not in ["LEFT", "RIGHT"]:
        fatal_error(f'{h_pos} is not valid, must be "left" or "right"!')

    # get the sizes of the images
    # subtract 1 from x and y since python counts start from 0
    if y != 0:
        y = y - 1
    if x != 0:
        x = x - 1
    # Convert grayscale images to color
    ori_img = _grayscale_to_rgb(img)

    # Image shape
    ix, iy = np.shape(ori_img)[0:2]

    # Convert mask to grayscale if needed and get its shape
    if any([len(np.shape(mask)) > 2, len(np.unique(mask)) > 2]):
        fatal_error("Mask should be a binary image")
    mx, my = np.shape(mask)

    # resize the images so they are equal in size and centered
    if mx >= ix:
        r = mx - ix
        r1 = int(np.rint(r / 2.0))
        r2 = _crop_modulo(r1, r)
        mask = mask[r1:mx - r2, 0:my]
    if my >= iy:
        r = my - iy
        r1 = int(np.rint(r / 2.0))
        r2 = _crop_modulo(r1, r)
        mask = mask[0:mx, r1:my - r2]

    # New mask shape
    mx, my = np.shape(mask)
    # make vertical adjustments
    if v_pos.upper() == "TOP":
        maskv = _top_crop_position(x, my, mask, ix)
    elif v_pos.upper() == "BOTTOM":
        maskv = _bottom_crop_position(x, my, mask, ix)
    # make horizontal adjustments
    if h_pos.upper() == "LEFT":
        maskv = _left_crop_position(y, maskv, iy)
    elif h_pos.upper() == "RIGHT":
        maskv = _right_crop_position(y, maskv, iy)

    newmask = np.array(maskv)
    _debug(
        visual=newmask,
        filename=os.path.join(params.debug_outdir, str(params.device) + "_newmask.png"),
        cmap="gray",
    )

    objects, hierarchy = _cv2_findcontours(bin_img=newmask)
    for i, _ in enumerate(objects):
        cv2.drawContours(
            ori_img, objects, i, (255, 102, 255), -1, lineType=8, hierarchy=hierarchy
        )
    _debug(
        visual=ori_img,
        filename=os.path.join(
            params.debug_outdir, str(params.device) + "_mask_overlay.png"
        ),
    )

    return newmask


def _crop_modulo(x1, x):
    """Force a number to make a rectangle slice

    Parameters
    ----------
    x1:       int, half the difference in size between two images
    x:        int, the difference in size between two images

    Returns
    -------
    out:      int, x1 or x1 - 1 depending on if the difference in size is even.
    """
    out = x1 if x % 2 == 0 else x1 - 1
    return out


def _top_crop_position(x, my, mask, ix):
    """Add rows to top of a mask to push it "down" on an image

    Parameters
    ----------
    x:       int, x position.
    my:      int, y-axis shape of image mask
    mask:    numpy.ndarray, binary mask
    ix:      int, x-axis shape of original image

    Returns
    -------
    maskv:   numpy.ndarray, binary mask moved vertically on original image scale.
    """
    # Add rows to the top
    top = np.zeros((x, my), dtype=np.uint8)

    maskv = np.vstack((top, mask))

    mx, my = np.shape(maskv)

    if mx >= ix:
        maskv = maskv[0:ix, 0:my]

    if mx < ix:
        r = ix - mx
        if r % 2 == 0:
            r1 = int(r / 2.0)
            rows1 = np.zeros((r1, my), dtype=np.uint8)
            maskv = np.vstack((rows1, maskv, rows1))
        else:
            r1 = int(math.ceil(r / 2.0))
            r2 = r1 - 1
            rows1 = np.zeros((r1, my), dtype=np.uint8)
            rows2 = np.zeros((r2, my), dtype=np.uint8)
            maskv = np.vstack((rows1, maskv, rows2))
    _debug(
        visual=maskv,
        filename=os.path.join(
            params.debug_outdir, str(params.device) + "_push-top.png"
        ),
        cmap="gray",
    )

    return maskv


def _bottom_crop_position(x, my, mask, ix):
    """Add rows to bottom of a mask to push it "up" on an image

    Parameters
    ----------
    x:       int, x position.
    my:      int, y-axis shape of image mask
    mask:    numpy.ndarray, binary mask
    ix:      int, x-axis shape of original image

    Returns
    -------
    maskv:   numpy.ndarray, binary mask moved vertically on original image scale.
    """
    # Add rows to the bottom
    bottom = np.zeros((x, my), dtype=np.uint8)

    maskv = np.vstack((mask, bottom))

    mx, my = np.shape(maskv)

    if mx >= ix:
        maskdiff = mx - ix
        maskv = maskv[maskdiff:mx, 0:my]

    if mx < ix:
        r = ix - mx
        if r % 2 == 0:
            r1 = int(r / 2.0)
            rows1 = np.zeros((r1, my), dtype=np.uint8)
            maskv = np.vstack((rows1, maskv, rows1))
        else:
            r1 = int(math.ceil(r / 2.0))
            r2 = r1 - 1
            rows1 = np.zeros((r1, my), dtype=np.uint8)
            rows2 = np.zeros((r2, my), dtype=np.uint8)
            maskv = np.vstack((rows1, maskv, rows2))

    _debug(
        visual=maskv,
        filename=os.path.join(
            params.debug_outdir, str(params.device) + "_push-bottom.png"
        ),
        cmap="gray",
    )

    return maskv


def _left_crop_position(y, mask, iy):
    """Add rows to bottom of a mask to push it "up" on an image

    Parameters
    ----------
    y:        int, y position for left adjustment.
    mask:     numpy.ndarray, binary mask after vertical adjustments.
    iy:       int, y-axis shape of original image.

    Returns
    -------
    maskh:   numpy.ndarray, binary mask moved vertically on original image scale.
    """
    mvx, _ = np.shape(mask)

    # Add rows to the left
    left = np.zeros((mvx, y), dtype=np.uint8)
    maskh = np.hstack((left, mask))

    mx, my = np.shape(maskh)

    if my >= iy:
        maskh = maskh[0:mx, 0:iy]
    if my < iy:
        c = iy - my
        if c % 2 == 0:
            c1 = int(c / 2.0)
            col = np.zeros((mx, c1), dtype=np.uint8)
            maskh = np.hstack((col, maskh, col))
        else:
            c1 = int(math.ceil(c / 2.0))
            c2 = c1 - 1
            col1 = np.zeros((mx, c1), dtype=np.uint8)
            col2 = np.zeros((mx, c2), dtype=np.uint8)
            maskh = np.hstack((col1, maskh, col2))

    _debug(
        visual=maskh,
        filename=os.path.join(
            params.debug_outdir, str(params.device) + "_push-left.png"
        ),
        cmap="gray",
    )

    return maskh


def _right_crop_position(y, mask, iy):
    """Add rows to bottom of a mask to push it "up" on an image

    Parameters
    ----------
    y:        int, y position for left adjustment.
    mask:     numpy.ndarray, binary mask after vertical adjustments.
    iy:       int, y-axis shape of original image.

    Returns
    -------
    maskh:   numpy.ndarray, binary mask moved vertically on original image scale.
    """
    mvx, _ = np.shape(mask)

    # Add rows to the left
    right = np.zeros((mvx, y), dtype=np.uint8)
    maskh = np.hstack((mask, right))

    mx, my = np.shape(maskh)

    if my >= iy:
        ex = my - iy
        maskh = maskh[0:mx, ex:my]

    if my < iy:
        c = iy - my
        if c % 2 == 0:
            c1 = int(c / 2.0)
            col = np.zeros((mx, c1), dtype=np.uint8)
            maskh = np.hstack((col, maskh, col))
        else:
            c1 = int(math.ceil(c / 2.0))
            c2 = c1 - 1
            col1 = np.zeros((mx, c1), dtype=np.uint8)
            col2 = np.zeros((mx, c2), dtype=np.uint8)
            maskh = np.hstack((col1, maskh, col2))

    _debug(
        visual=maskh,
        filename=os.path.join(
            params.debug_outdir, str(params.device) + "_push-right.png"
        ),
        cmap="gray",
    )

    return maskh



"""Analyzes the shape and size of objects and outputs data."""
import os
import cv2
import numpy as np
from scipy.spatial.distance import euclidean
from plantcv.plantcv._helpers import _iterate_analysis, _cv2_findcontours, _object_composition, _grayscale_to_rgb, _scale_size
from plantcv.plantcv import outputs, within_frame
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug


def size(img, labeled_mask, n_labels=1, label=None):
    """A function that analyzes the shape and size of objects and outputs data.

    Inputs:
    img          = RGB or grayscale image data for plotting
    labeled_mask = Labeled mask of objects (32-bit).
    n_labels     = Total number expected individual objects (default = 1).
    label        = Optional label parameter, modifies the variable name of
                   observations recorded (default = pcv.params.sample_label).

    Returns:
    analysis_image = Diagnostic image showing measurements.

    :param img: numpy.ndarray
    :param labeled_mask: numpy.ndarray
    :param n_labels: int
    :param label: str
    :return analysis_image: numpy.ndarray
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    img = _iterate_analysis(img=img, labeled_mask=labeled_mask, n_labels=n_labels, label=label, function=_analyze_size)
    # Debugging
    _debug(visual=img, filename=os.path.join(params.debug_outdir, str(params.device) + '_shapes.png'))
    return img


def _analyze_size(img, mask, label):
    """Analyze the size of individual objects.

    Inputs:
    img   = RGB or grayscale image data for plotting
    mask  = Binary image data
    label = Label of object

    Returns:
    analysis_image = Diagnostic image showing measurements

    :param mask: numpy.ndarray
    :param label: int
    :return analysis_image: numpy.ndarray
    """
    params.device += 1
    # Initialize analysis output values
    area = 0
    hull_area = 0
    solidity = 0
    perimeter = 0
    total_edge_length = 0
    width = 0
    height = 0
    caliper_length = 0
    longest_path = 0
    cmx = 0
    cmy = 0
    hull_vertices = 0
    ellipse_center = 0, 0
    ellipse_major_axis = 0
    ellipse_minor_axis = 0
    ellipse_angle = 0
    ellipse_eccentricity = 0
    # Check is object is touching image boundaries (QC)
    in_bounds = within_frame(mask=mask, label=label)

    # Convert grayscale images to color
    img = _grayscale_to_rgb(img)
    # Plot image
    plt_img = np.copy(img)

    # Find contours
    cnt, cnt_str = _cv2_findcontours(bin_img=mask)

    # Consolidate contours
    obj = _object_composition(contours=cnt, hierarchy=cnt_str)

    # Analyze shape properties if the object is large enough
    if len(obj) > 5:
        # Convex Hull
        hull = cv2.convexHull(obj)
        # Number of convex hull vertices
        hull_vertices = len(hull)
        # Convex Hull area
        hull_area = cv2.contourArea(hull)
        # Moments
        m = cv2.moments(mask, binaryImage=True)
        # Area
        area = m['m00']
        # Solidity
        solidity = area / hull_area if hull_area != 0 else 1
        # Perimeter
        perimeter = cv2.arcLength(obj, closed=True)
        # Total edge legnth
        for contour in cnt:
            total_edge_length += cv2.arcLength(contour, True)
        # Bounding rectangle
        x, y, width, height = cv2.boundingRect(obj)
        # Centroid/Center of Mass
        cmx = m['m10'] / m['m00']
        cmy = m['m01'] / m['m00']
        # Bounding ellipse
        ellipse_center, axes, ellipse_angle = cv2.fitEllipse(obj)
        major_axis_idx = np.argmax(axes)
        minor_axis_idx = 1 - major_axis_idx
        ellipse_major_axis = float(axes[major_axis_idx])
        ellipse_minor_axis = float(axes[minor_axis_idx])
        ellipse_eccentricity = float(np.sqrt(1 - (axes[minor_axis_idx] / axes[major_axis_idx]) ** 2))
        # Caliper length
        caliper_length, caliper_transpose = _longest_axis(height=img.shape[0], width=img.shape[1],
                                                          hull=hull, cmx=cmx, cmy=cmy)
        longest_path = euclidean(tuple(caliper_transpose[caliper_length - 1]), tuple(caliper_transpose[0]))

        # Add measurements onto the diagnostic image
        # color blind friendly palette in BGR: (255, 0, 255) = magenta; (255, 0, 0) = blue
        # Draw convex hull
        cv2.drawContours(plt_img, [hull], -1, (255, 0, 255), params.line_thickness)
        # Draw perimeter outline
        cv2.drawContours(plt_img, cnt, -1, (255, 0, 0), params.line_thickness)
        # Draw width
        cv2.line(plt_img, (x, y), (x + width, y), (255, 0, 255), params.line_thickness)
        # Draw height
        cv2.line(plt_img, (int(cmx), y), (int(cmx), y + height), (255, 0, 255), params.line_thickness)
        # Draw centroid
        cv2.circle(plt_img, (int(cmx), int(cmy)), 10, (255, 0, 255), params.line_thickness)
        # Draw longest path
        cv2.line(plt_img, (tuple(caliper_transpose[caliper_length - 1])), (tuple(caliper_transpose[0])),
                 (255, 0, 255), params.line_thickness)
        if params.verbose:
            # Label the object with object label
            cv2.putText(img=plt_img, text=label, org=(int(cmx), int(cmy)), fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                        fontScale=params.text_size, color=(187, 187, 187), thickness=params.text_thickness)

    # Store outputs
    outputs.add_metadata(term="image_height", datatype=int, value=np.shape(img)[0])
    outputs.add_metadata(term="image_width", datatype=int, value=np.shape(img)[1])
    outputs.add_observation(sample=label, variable='area', trait='area',
                            method='plantcv.plantcv.analyze.size', scale=params.unit, datatype=int,
                            value=_scale_size(area, "area"), label=params.unit)
    outputs.add_observation(sample=label, variable='convex_hull_area', trait='convex hull area',
                            method='plantcv.plantcv.analyze.size', scale=params.unit, datatype=int,
                            value=_scale_size(hull_area, "area"), label=params.unit)
    outputs.add_observation(sample=label, variable='solidity', trait='solidity',
                            method='plantcv.plantcv.analyze.size', scale='none', datatype=float,
                            value=solidity, label='none')
    outputs.add_observation(sample=label, variable='perimeter', trait='perimeter',
                            method='plantcv.plantcv.analyze.size', scale=params.unit, datatype=int,
                            value=_scale_size(perimeter), label=params.unit)
    outputs.add_observation(sample=label, variable='total_edge_length', trait='total length of object edges',
                            method='plantcv.plantcv.analyze.size', scale=params.unit, datatype=int,
                            value=_scale_size(total_edge_length), label=params.unit)
    outputs.add_observation(sample=label, variable='width', trait='width',
                            method='plantcv.plantcv.analyze.size', scale=params.unit, datatype=int,
                            value=_scale_size(width), label=params.unit)
    outputs.add_observation(sample=label, variable='height', trait='height',
                            method='plantcv.plantcv.analyze.size', scale=params.unit, datatype=int,
                            value=_scale_size(height), label=params.unit)
    outputs.add_observation(sample=label, variable='longest_path', trait='longest path',
                            method='plantcv.plantcv.analyze.size', scale=params.unit, datatype=int,
                            value=_scale_size(float(longest_path)), label=params.unit)
    outputs.add_observation(sample=label, variable='center_of_mass', trait='center of mass',
                            method='plantcv.plantcv.analyze.size', scale='none', datatype=tuple,
                            value=(cmx, cmy), label=("x", "y"))
    outputs.add_observation(sample=label, variable='convex_hull_vertices', trait='convex hull vertices',
                            method='plantcv.plantcv.analyze.size', scale='none', datatype=int,
                            value=hull_vertices, label='none')
    outputs.add_observation(sample=label, variable='object_in_frame', trait='object in frame',
                            method='plantcv.plantcv.analyze.size', scale='none', datatype=bool,
                            value=in_bounds, label='none')
    outputs.add_observation(sample=label, variable='ellipse_center', trait='ellipse center',
                            method='plantcv.plantcv.analyze.size', scale='none', datatype=tuple,
                            value=(ellipse_center[0], ellipse_center[1]), label=("x", "y"))
    outputs.add_observation(sample=label, variable='ellipse_major_axis', trait='ellipse major axis length',
                            method='plantcv.plantcv.analyze.size', scale=params.unit, datatype=int,
                            value=_scale_size(ellipse_major_axis), label=params.unit)
    outputs.add_observation(sample=label, variable='ellipse_minor_axis', trait='ellipse minor axis length',
                            method='plantcv.plantcv.analyze.size', scale=params.unit, datatype=int,
                            value=_scale_size(ellipse_minor_axis), label=params.unit)
    outputs.add_observation(sample=label, variable='ellipse_angle', trait='ellipse major axis angle',
                            method='plantcv.plantcv.analyze.size', scale='degrees', datatype=float,
                            value=float(ellipse_angle), label='degrees')
    outputs.add_observation(sample=label, variable='ellipse_eccentricity', trait='ellipse eccentricity',
                            method='plantcv.plantcv.analyze.size', scale='none', datatype=float,
                            value=float(ellipse_eccentricity), label='none')
    return plt_img


def _longest_axis(height, width, hull, cmx, cmy):
    """
    Calculate the line through center of mass and point on the convex hull that is furthest away

    :param height: int
    :param width: int
    :param hull: numpy.ndarray
    :param cmx: int
    :param cmy: int
    :return caliper_length: int
    """
    background = np.zeros((height, width, 3), np.uint8)
    background1 = np.zeros((height, width), np.uint8)
    background2 = np.zeros((height, width), np.uint8)
    # Longest Axis: line through center of mass and point on the convex hull that is furthest away
    cv2.circle(background, (int(cmx), int(cmy)), 4, (255, 255, 255), -1)
    center_p = cv2.cvtColor(background, cv2.COLOR_BGR2GRAY)
    _, centerp_binary = cv2.threshold(center_p, 0, 255, cv2.THRESH_BINARY)
    centerpoint, _ = _cv2_findcontours(bin_img=centerp_binary)

    dist = []
    vhull = np.vstack(hull)

    for i, c in enumerate(vhull):
        xy = tuple(int(ci) for ci in c)
        pptest = cv2.pointPolygonTest(centerpoint[0], xy, measureDist=True)
        dist.append(pptest)

    abs_dist = np.absolute(dist)
    max_i = np.argmax(abs_dist)

    caliper_max_x, caliper_max_y = list(tuple(vhull[max_i]))
    caliper_mid_x, caliper_mid_y = [int(cmx), int(cmy)]

    xdiff = float(caliper_max_x - caliper_mid_x)
    ydiff = float(caliper_max_y - caliper_mid_y)

    # Set default values
    slope = 1

    if xdiff != 0:
        slope = float(ydiff / xdiff)
    b_line = caliper_mid_y - (slope * caliper_mid_x)

    if slope != 0:
        xintercept = int(-b_line / slope)
        xintercept1 = int((height - b_line) / slope)
        if 0 <= xintercept <= width and 0 <= xintercept1 <= width:
            cv2.line(background1, (xintercept1, height), (xintercept, 0), (255), params.line_thickness)
        elif xintercept < 0 or xintercept > width or xintercept1 < 0 or xintercept1 > width:
            yintercept = int(b_line)
            yintercept1 = int((slope * width) + b_line)
            cv2.line(background1, (0, yintercept), (width, yintercept1), (255), 5)
    else:
        cv2.line(background1, (width, caliper_mid_y), (0, caliper_mid_y), (255), params.line_thickness)

    _, line_binary = cv2.threshold(background1, 0, 255, cv2.THRESH_BINARY)

    cv2.drawContours(background2, [hull], -1, (255), -1)
    _, hullp_binary = cv2.threshold(background2, 0, 255, cv2.THRESH_BINARY)

    caliper = cv2.multiply(line_binary, hullp_binary)

    caliper_y, caliper_x = np.array(caliper.nonzero())
    caliper_matrix = np.vstack((caliper_x, caliper_y))
    caliper_transpose = np.transpose(caliper_matrix)
    caliper_length = len(caliper_transpose)

    caliper_transpose1 = np.lexsort((caliper_y, caliper_x))
    caliper_transpose2 = [(caliper_x[i], caliper_y[i]) for i in caliper_transpose1]
    caliper_transpose = np.array(caliper_transpose2)

    return caliper_length, caliper_transpose



import os
import altair as alt
import pandas as pd
import numpy as np
from plantcv.plantcv import params, fatal_error, PSII_data
from plantcv.plantcv._debug import _debug


def chlorophyll_fluorescence(ps_da, labeled_mask, n_labels=1, label="object"):
    """Plot the chlorophyll fluorescence induction curve for each object.

    Inputs:
    ps_da            = photosynthesis xarray DataArray
    labeled_mask     = Labeled mask of objects (32-bit).
    n_labels         = Total number expected individual objects (default = 1).
    label            = optional label parameter, modifies the prefix of the group plotting label

    Returns:
    chart            = Plot of the chlorophyll fluorescence induction curve for each object

    :param ps_da: xarray.DataArray
    :param labeled_mask: numpy.ndarray
    :param n_labels: int
    :param label: str
    :return chart: altair.Chart
    """
    # Increment function counter
    params.device += 1

    # Check that the dataarray is valid
    try:
        if ps_da.name not in ["ojip_light", "ojip_dark"]:
            fatal_error("You must provide a xarray DataArray with name ojip_light or ojip_dark")
    except AttributeError:
        if isinstance(ps_da, PSII_data):
            fatal_error("You need to provide the `ojip_dark` or `ojip_light` dataarray")
        else:
            fatal_error("You must provide a xarray DataArray with name ojip_light or ojip_dark")

    # Prime is empty for Fv/Fm (dark- and light-adapted) and p for Fq'/Fm'
    datasets = {
        "ojip_light": {
            "prime": "p",
            "label": "PSL"
        },
        "ojip_dark": {
            "prime": "",
            "label": "PSD"
        }
    }

    # Get the number of frame labels
    ind_size = ps_da.frame_label.size

    # Find the current Fm or Fm'
    idx = list(ps_da.frame_label.values).index(f"Fm{datasets[ps_da.name.lower()]['prime']}")

    # Initialize a dictionary to store the induction curve data
    data = {"Timepoints": [], "Fluorescence": [], "Labels": [], "Group": []}

    mask_copy = np.copy(labeled_mask)
    if len(np.unique(mask_copy)) == 2 and np.max(mask_copy) == 255:
        mask_copy = np.where(mask_copy == 255, 1, 0).astype(np.uint8)
    for i in range(1, n_labels + 1):
        # Create a boolean submask for each label
        submask = np.where(mask_copy == i, True, False).astype(bool)

        # Get plant mean for each frame based on mask
        fluor_values = ps_da.where(submask[..., None, None]).mean(['x', 'y', 'measurement']).values

        # Append fluorescence values to data dictionary
        data["Timepoints"].extend(range(0, ind_size))
        data["Fluorescence"].extend(list(fluor_values))
        data["Labels"].extend(list(ps_da.frame_label.values))
        data["Group"].extend([f"{label}{i}"] * ind_size)

    # Create a dataframe from the data dictionary
    df = pd.DataFrame(data)

    # Create a chart
    chart = (
        alt.Chart(df, title=f"{ps_da.name} induction curve")
        .mark_line(point=True)
        .encode(
            x="Timepoints:Q",
            y="Fluorescence:Q",
            color="Group:N",
            tooltip=["Group", "Timepoints", "Fluorescence"]).interactive()
        )
    # Create a vertical rule for the Fm or Fm' value
    rule = (
        alt.Chart(df)
        .mark_rule(strokeDash=[10, 10], color="gray")
        .encode(x=alt.datum(idx))
    )
    # Label the chart with the Fm or Fm' value
    text = (
        alt.Chart(df.query(f"Labels == 'Fm{datasets[ps_da.name.lower()]['prime']}'").iloc[[0]])
        .mark_text(dy=-15)
        .encode(x="Timepoints:Q", y="Fluorescence:Q", text="Labels:N")
    )

    # Plot debug image
    _debug(visual=chart + text + rule,
           filename=os.path.join(params.debug_outdir, f"{params.device}_fluorescence_plot.png"))

    return chart + text + rule



"""Plot Empirical Cumulative Distribution Function for Object Size."""
import os
import cv2
import pandas as pd
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _cv2_findcontours
from statsmodels.distributions.empirical_distribution import ECDF
import altair as alt


def obj_size_ecdf(mask):
    """
    Plot empirical cumulative distribution for object size based on binary mask.

    Inputs:
    mask  = binary mask

    Returns:
    chart = empirical cumulative distribution function plot

    :param mask: numpy.ndarray
    :return chart: altair.vegalite.v5.api.Chart
    """
    objects, _ = _cv2_findcontours(bin_img=mask)
    areas = [cv2.contourArea(cnt) for cnt in objects]
    # Remove objects with areas < 1px
    areas = [i for i in areas if i >= 1.0]

    ecdf = ECDF(areas, side='right')

    ecdf_df = pd.DataFrame({'object area': ecdf.x[1:], 'cumulative probability': ecdf.y[1:]})
    # create ecdf plot and apply log-scale for x-axis (areas)
    chart = alt.Chart(ecdf_df).mark_circle(size=10).encode(
        x=alt.X("object area:Q").scale(type='log'),
        y="cumulative probability:Q",
        tooltip=['object area', 'cumulative probability']
    )

    # Plot or print the ecdf
    _debug(visual=chart,
           filename=os.path.join(params.debug_outdir, str(params.device) + '_area_ecdf.png'))
    return chart.interactive()



# Read image with bayer mosaic

import os
import cv2
from plantcv.plantcv import fatal_error
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import params


def readbayer(filename, bayerpattern='BG', alg='default'):
    """Read image from file that has a Bayer mosaic.

    Parameters
    ----------
    filename : str
        Name of image file
    bayerpattern : str, optional
        Arrangement of the pixels. ("BG","GB","RG","GR"), by default 'BG'
    alg : str, optional
        Algorithm with which to demosaic the image ("default","EdgeAware","VariableNumberGradients"), by default 'default'

    Returns
    -------
    numpy.ndarray
        Image with Bayer mosaic demosaiced
    str
        Path to the image file
    str
        Name of the image file
    """
    # bayerpattern is defined as the colors of the pixels in the 2nd and 3rd column of the 2nd row.
    # see https://docs.opencv.org/3.2.0/de/d25/imgproc_color_conversions.html
    # COLOR_BayerBG2BGR
    # COLOR_BayerGB2BGR
    # COLOR_BayerRG2BGR
    # COLOR_BayerGR2BGR
    image_raw = cv2.imread(filename, flags=-1)

    if image_raw is None:
        fatal_error(f"Failed to open {filename}")

    # Itemize the Bayer pattern and algorithm
    converters = {
        "DEFAULT": {
            "BG": cv2.COLOR_BayerBG2BGR,
            "GB": cv2.COLOR_BayerGB2BGR,
            "RG": cv2.COLOR_BayerRG2BGR,
            "GR": cv2.COLOR_BayerGR2BGR
        },
        "EDGEAWARE": {
            "BG": cv2.COLOR_BayerBG2BGR_EA,
            "GB": cv2.COLOR_BayerGB2BGR_EA,
            "RG": cv2.COLOR_BayerRG2BGR_EA,
            "GR": cv2.COLOR_BayerGR2BGR_EA
        },
        "VARIABLENUMBERGRADIENTS": {
            "BG": cv2.COLOR_BayerBG2BGR_VNG,
            "GB": cv2.COLOR_BayerGB2BGR_VNG,
            "RG": cv2.COLOR_BayerRG2BGR_VNG,
            "GR": cv2.COLOR_BayerGR2BGR_VNG
        }
    }

    # Check if the algorithm is valid
    if alg.upper() not in converters:
        fatal_error(f"Invalid algorithm '{alg}'. Choose from {list(converters.keys())}.")
    # Check if the bayer pattern is valid
    if bayerpattern.upper() not in converters[alg.upper()]:
        fatal_error(f"Invalid Bayer pattern '{bayerpattern}'. Choose from {list(converters[alg.upper()].keys())}.")
    # Convert the image using the specified algorithm and Bayer pattern
    img = cv2.cvtColor(image_raw, converters[alg.upper()][bayerpattern.upper()])

    # Split path from filename
    path, img_name = os.path.split(filename)

    _debug(visual=img,
           filename=os.path.join(params.debug_outdir, "input_image.png"))

    return img, path, img_name



# Object fill device

import numpy as np
import os
from plantcv.plantcv import fatal_error
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _rect_filter, _rect_replace
from skimage.morphology import remove_small_objects


def fill(bin_img, size, roi=None):
    """
    Identifies objects and fills objects that are less than size.

    Inputs:
    bin_img      = Binary image data
    size         = minimum object area size in pixels (integer)
    roi          = optional Objects class rectangular ROI

    Returns:
    filtered_img = image with objects filled

    :param bin_img: numpy.ndarray
    :param size: int
    :param roi: plantcv.plantcv.Objects
    :return filtered_img: numpy.ndarray
    """
    # Make sure the image is binary
    if len(np.shape(bin_img)) != 2 or len(np.unique(bin_img)) > 2:
        fatal_error("Image is not binary")

    # Cast binary image to boolean
    bool_img = bin_img.astype(bool)

    # Find and fill contours, possibly within bounding rectangle
    bool_img = _rect_filter(bool_img,
                            roi=roi,
                            function=remove_small_objects,
                            **{"min_size" : size})
    # Cast boolean image to binary and make a copy of the binary image for returning
    filtered_img = np.copy(bool_img.astype(np.uint8) * 255)
    # slice the subset image back into full size binary image
    replaced_img = _rect_replace(bin_img.astype(bool) * 255, filtered_img, roi)

    _debug(visual=replaced_img,
           filename=os.path.join(params.debug_outdir, str(params.device) + "_fill" + str(size) + '.png'))

    return replaced_img



# Point/vertice annotation tool(s)

import numpy as np
from scipy.spatial import distance

# INTERACTIVE ROI TOOLS #


def _find_closest_pt(pt, pts):
    """Given coordinates of a point and a list of coordinates of a bunch of points, find the point that has the
    smallest Euclidean to the given point

    :param pt: (tuple) coordinates of a point
    :param pts: (a list of tuples) coordinates of a list of points
    :return: index of the closest point and the coordinates of that point
    """
    if pt in pts:
        idx = pts.index(pt)
        return idx, pt

    dists = distance.cdist([pt], pts, 'euclidean')
    idx = np.argmin(dists)
    return idx, pts[idx]



# Classify pixels as plant or non-plant using the naive Bayes method written by Arash Abbasi,
# adapted for Python by Noah Fahlgren

import cv2
import numpy as np
import os
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import fatal_error
from plantcv.plantcv import params


def naive_bayes_classifier(rgb_img, pdf_file):
    """
    Use the Naive Bayes classifier to output a plant binary mask.

    Inputs:
    rgb_img      = RGB image data
    pdf_file = filename of file containing PDFs output from the Naive Bayes training method (see plantcv-train.py)

    Returns:
    mask     = Dictionary of binary masks

    :param rgb_img: numpy.ndarray
    :param pdf_file: str
    :return masks: dict
    """
    # Initialize PDF dictionary
    pdfs = {}
    # Read the PDF file
    pf = open(pdf_file, "r")
    # Read the first line (header)
    pf.readline()
    # Read each line of the file and parse the PDFs, store in the PDF dictionary
    for row in pf:
        # Remove newline character
        row = row.rstrip("\n")
        # Split the row into columns on tab characters
        cols = row.split("\t")
        # Make sure there are the correct number of columns (i.e. is this a valid PDF file?)
        if len(cols) != 258:
            fatal_error("Naive Bayes PDF file is not formatted correctly. Error on line:\n" + row)
        # Store the PDFs. Column 0 is the class, Column 1 is the color channel, the rest are p at
        # intensity values 0-255. Cast text p values as float
        class_name = cols[0]
        channel = cols[1]
        if class_name not in pdfs:
            pdfs[class_name] = {}
        pdfs[class_name][channel] = np.array([float(i) for i in cols[2:]])

    # Split the input BGR image into component channels for BGR, HSV, and LAB colorspaces
    h, s, v = cv2.split(cv2.cvtColor(rgb_img, cv2.COLOR_BGR2HSV))

    # Calculate the dimensions of the input image
    width, height, _ = np.shape(rgb_img)

    # Linearize the color channel arrays
    hIdx = h.reshape(width * height).tolist()
    sIdx = s.reshape(width * height).tolist()
    vIdx = v.reshape(width * height).tolist()

    # Calculate the joint probability that this is in the class
    class_arrs = []
    for class_name in pdfs:
        # Joint probability array
        arr = pdfs[class_name]["hue"][hIdx] * \
            pdfs[class_name]["saturation"][sIdx] * \
            pdfs[class_name]["value"][vIdx]
        # Append the joint probability array to the list of class arrays
        class_arrs.append(arr.reshape(width, height))

    # Stacked class arrays
    stacked_arr = np.dstack(class_arrs)
    # Create an array that labels which class has the highest probability
    class_mask = np.argmax(stacked_arr, axis=2)

    # Create the class masks
    masks = {}
    for i, class_name in enumerate(pdfs):
        # Set pixel intensities to 255 (white) where the class has the highest probability
        masks[class_name] = np.where(class_mask == i, 255, 0).astype(np.uint8)

    # Print or plot the mask if debug is not None
    for class_name, mask in masks.items():
        _debug(visual=mask, filename=os.path.join(params.debug_outdir, f"{params.device}_naive_bayes_{class_name}_mask.png"),
               cmap='gray')

    return masks



# Automatically detect a color card and color correct to standard chip values

from plantcv.plantcv import params, deprecation_warning
from plantcv.plantcv.transform.detect_color_card import detect_color_card
from plantcv.plantcv.transform.color_correction import get_color_matrix, std_color_matrix, affine_color_correction


def auto_correct_color(rgb_img, label=None, color_chip_size=None, roi=None, **kwargs):
    """Automatically detect a color card.
    Parameters
    ----------
    rgb_img : numpy.ndarray
        Input RGB image data containing a color card.
    label : str, optional
        modifies the variable name of observations recorded (default = pcv.params.sample_label).
    color_chip_size: str, tuple, optional
        "passport", "classic", "cameratrax"; or tuple formatted (width, height)
        in millimeters (default = None)
    roi: plantcv.plantcv.Objects, optional
        Objects class rectangular ROI passed to detect_color_card (default None)
    **kwargs
        Other keyword arguments passed to cv2.adaptiveThreshold, cv2.circle and _rect_filter.
        Valid keyword arguments:
        adaptive_method: 0 (mean) or 1 (Gaussian) (default = 1)
        block_size: int (default = 51)
        radius: int (default = 20)
        min_size: int (default = 1000)
        aspect_ratio: float (default = 1.27)
        solidity: float (default = 0.8)
    Returns
    -------
    numpy.ndarray
        Color corrected image
    """
    # Set label to params.sample_label if None
    if label is None:
        label = params.sample_label
    deprecation_warning(
        "The 'label' parameter is no longer utilized, since color chip size is now metadata. "
        "It will be removed in PlantCV v5.0."
        )
    labeled_mask = detect_color_card(rgb_img=rgb_img, color_chip_size=color_chip_size, roi=roi, **kwargs)
    _, card_matrix = get_color_matrix(rgb_img=rgb_img, mask=labeled_mask)
    std_matrix = std_color_matrix(pos=3)
    return affine_color_correction(rgb_img=rgb_img, source_matrix=card_matrix,
                                   target_matrix=std_matrix)



# Calculate the inverse covariance matrix of a hyperspectral datacube

import numpy as np


def _inverse_covariance(spectral_array):
    """Calculate the inverse covariance matrix of a hyperspectral datacube.
    Used in various GatorSense hyperspectral tools (https://github.com/GatorSense/hsi_toolkit_py)

    Inputs:
    spectral_array      = Hyperspectral data instance

    Returns:
    inverse_covariance  = Inverse covariance matrix of a hyperspectral datacube

    :param spectral_array: __main__.Spectral_data
    :return inverse_covariance: numpy array
    """
    hsi_img = spectral_array.array_data

    n_lines, n_samples, n_band = hsi_img.shape
    n_pixels = n_lines * n_samples
    hsi_data = np.reshape(hsi_img, (n_pixels, n_band), order='F').T
    inverse_covariance = np.linalg.pinv(np.cov(hsi_data.T, rowvar=False))

    return inverse_covariance



# Find both segment end coordinates
import os
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _find_segment_ends


def segment_ends(skel_img, leaf_objects, mask=None):
    """Find tips and segment branch points, resort objects based on inner_list of coordinates.

    Inputs:
    skel_img         = Skeletonized image
    leaf_objects     = List of leaf segments
    mask             = (Optional) binary mask for debugging. If provided, debug image will be overlaid on the mask.

    Returns:
    sorted_obs       = Reordered segments based on segment branch point y-coordinates
    inner_list       = List of branch point coordinates of the input leaf_objects
    tip_list         = List of leaf tip coordinates of the input leaf_object

    :param segmented_img: numpy.ndarray
    :param leaf_objects: list
    :param mask: numpy.ndarray
    :return sorted_ids: list
    :return inner_list: list
    :return tip_list: list
    """
    # Store debug
    debug = params.debug
    params.debug = None

    if mask is None:
        labeled_img = skel_img.copy()
    else:
        labeled_img = mask.copy()
    # Find and sort segment ends, and create debug image
    labeled_img, tip_list, inner_list, _, objs = _find_segment_ends(
        skel_img=skel_img, leaf_objects=leaf_objects, plotting_img=labeled_img, size=1)

    # Reset debug mode
    params.debug = debug
    _debug(visual=labeled_img, filename=os.path.join(params.debug_outdir, f"{params.device}_segment_ends.png"))

    # Determine optimal segment order by y-coordinate order
    d = {}
    for i, coord in enumerate(inner_list):
        d[i] = coord[1]  # y-coord is the key and index the value
    values = list(d.values())
    sorted_key_index = np.argsort(values)
    sorted_objs = [objs[i] for i in sorted_key_index[::-1]]

    return sorted_objs, inner_list, tip_list



# training a patch-based kmeans clustering model on a set of images

import os
import numpy as np
import cv2
import random
from sklearn.cluster import MiniBatchKMeans
from sklearn.feature_extraction import image
from skimage.filters import gaussian
from joblib import dump


def train_kmeans(img_dir, k, out_path="./kmeansout.fit", prefix="", patch_size=10, sigma=5, sampling=None,
                 seed=1, num_imgs=0, n_init=10):
    """
    Trains a patch-based kmeans clustering model for identifying image features.
    Inputs:
    img_idr = Path to directory where training images are stored
    K = Number of clusters to fit
    out_path = Path to directory where the model output should be stored
    prefix = Keyword for target images. Anything in img_dir without the prefix will be skipped
    patch_size = Size of the NxN neighborhood around each pixel
    sigma = Gaussian blur sigma. Denotes severity of gaussian blur performed before patch identification
    sampling = Fraction of image from which patches are identified
    seed = Seed for determinism of random elements like sampling of patches
    num_imgs = Number of images to use for training. Default is all of them in img_dir with prefix
    n_init = Number of random initiations tried by MiniBatchKMeans. The algorithm is run on the best one
    :param img_dir: str
    :param K: positive non-zero integer
    :param out_path: str
    :param prefix: str
    :param patch_size: positive non-zero integer
    :param sigma: positive real number or sequence of positive real numbers
    :param sampling: float (0,1]
    :param seed: positive integer
    :param num_imgs: positive non-zero integer
    :param n_init: positive non-zero integer
    :return fitted: sklearn.cluster._kmeans.MiniBatchKMeans
    """
    # Establish training set
    file_names = sorted(os.listdir(img_dir))
    if num_imgs == 0:
        training_files = file_names
    else:
        training_files = random.choices(file_names, k=num_imgs)  # choosing a set of random files
    # Read and extract patches
    for idx, img_name in enumerate(training_files):
        if prefix in img_name:
            img = cv2.imread(os.path.join(img_dir, img_name), -1)
            # Check to make sure non-image files in the directory are not included
            if isinstance(img, np.ndarray):
                if idx == 0:
                    # Getting info from first image
                    patches = patch_extract(img, patch_size=patch_size, sigma=sigma, sampling=sampling)
                else:
                    # Concatenating each additional image
                    patches = np.vstack((patches, patch_extract(img, patch_size=patch_size, sigma=sigma,
                                                                sampling=sampling)))
    kmeans = MiniBatchKMeans(n_clusters=k, n_init=n_init, random_state=seed)
    fitted = kmeans.fit(patches)
    dump(fitted, out_path)
    return fitted


def patch_extract(img, patch_size=10, sigma=5, sampling=None, seed=1):
    """
    Extracts patches from an image.
    Inputs:
    img = An image from which to extract patches
    patch_size = Size of the NxN neighborhood around each pixel
    sigma = sigma = Gaussian blur sigma. Denotes severity of gaussian blur performed before patch identification
    sampling = Fraction of image from which patches are identified
    seed = Seed for determinism of random elements like sampling of patches
    :param img: numpy.ndarray
    :param patch_size: positive non-zero integer
    :param sigma: positive real number or sequence of positive real numbers
    :param sampling: float (0,1]
    :param seed: positive integer
    :return patches_lin: numpy.ndarray
    """
    # Gaussian blur
    if len(img.shape) == 2:
        img_blur = np.round(gaussian(img, sigma=sigma)*255).astype(np.uint16)
    elif len(img.shape) == 3 and img.shape[2] == 3:
        img_blur = np.round(gaussian(img, sigma=sigma, channel_axis=2)*255).astype(np.uint16)

    # Extract patches
    patches = image.extract_patches_2d(img_blur, (patch_size, patch_size),
                                       max_patches=sampling, random_state=seed)
    N = patches.shape[0]
    patches_lin = patches.reshape(N, -1)
    return patches_lin



import os
import sys
import json
from copy import deepcopy
import uuid


# Build job list
###########################################
def job_builder(meta, config):
    """Build a list of image processing jobs.

    Inputs:
    meta:         Dictionary of processed image metadata.
    config:       plantcv.parallel.WorkflowConfig object.

    Returns:
    jobs:         List of image processing commands.

    :param meta: dict
    :param config: plantcv.parallel.WorkflowConfig
    :return job_stack: list
    """
    # Overall job stack. List of list of jobs
    jobs = []

    # Log the number of jobs to be run
    n_jobs = len(meta)
    print(f"Task list includes {n_jobs} workflows", file=sys.stderr)

    # Each grouping has a tuple of grouped metadata values and a dataframe of image metadata
    for _, grp_df in meta:
        # Create a JSON template for each group
        img_meta = {"metadata": deepcopy(config.metadata_terms), "observations": {}}

        # Store metadata in JSON
        img_meta["metadata"]["image"] = {
                "label": "image files",
                "datatype": "<class 'str'>",
                "value": grp_df["filepath"].values.tolist()
            }

        # Convert datetime to string before serialization
        grp_df["timestamp"] = grp_df["timestamp"].dt.strftime(config.timestampformat)

        # Valid metadata
        for m in list(config.metadata_terms.keys()):
            img_meta["metadata"][m]["value"] = grp_df[m].values.tolist()

        # Create random unique output file to store the image processing results and populate with metadata
        outfile = os.path.join(config.tmp_dir, f"{uuid.uuid4()}.json")
        with open(outfile, "w") as fp:
            json.dump(img_meta, fp, indent=4)

        # Generate image names if set to auto
        if config.group_name == "auto":
            names = []
            for i in range(0, len(grp_df)):
                names.append(f"image{i + 1}")
        else:
            names = list(grp_df[config.group_name])

        # Build job
        job_parts = ["python", config.workflow, "--outdir", config.img_outdir, "--result", outfile,
                     "--names", ",".join(map(str, names))]
        # Add other arguments
        for key, value in config.other_args.items():
            job_parts.append(f"--{key}")
            job_parts.append(value)
        if config.writeimg:
            job_parts.append("--writeimg")
        for fname in grp_df["filepath"].values.tolist():
            job_parts.append(fname)
        jobs.append(job_parts)

    return jobs
###########################################



"""Read in fluorescence images from a .DAT file."""
import os
import numpy as np
import xarray as xr
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import PSII_data
from plantcv.plantcv import Spectral_data
from skimage.util import img_as_ubyte


def read_cropreporter(filename):
    """
    Read datacubes from PhenoVation B.V. CropReporter into a PSII_Data instance.

    Inputs:
        filename        = CropReporter .INF filename

    Returns:
        ps               = photosynthesis data in xarray DataArray format

    :param filename: str
    :return ps: plantcv.plantcv.classes.PSII_data
    """
    # Initialize metadata dictionary
    metadata_dict = {}

    # Parse .inf file and create dictionary with metadata stored within
    with open(filename, "r") as fp:
        for line in fp:
            if "=" in line:
                key, value = line.rstrip("\n").split("=")
                metadata_dict[key] = value

    # Initialize PSII_data class
    ps = PSII_data()

    # INF file prefix and path
    ps.filename = os.path.split(filename)[-1]
    ps.datapath = os.path.dirname(filename)

    # Dark-adapted measurements
    _process_psd_data(ps=ps, metadata=metadata_dict)

    # Light-adapted measurements
    _process_psl_data(ps=ps, metadata=metadata_dict)

    # Dark-adapted PAM measurements
    _process_pmd_data(ps=ps, metadata=metadata_dict)

    # Light-adapted PAM Pmeasurements
    _process_pml_data(ps=ps, metadata=metadata_dict)

    # Chlorophyll fluorescence data
    _process_chl_data(ps=ps, metadata=metadata_dict)

    # Spectral measurements
    _process_spc_data(ps=ps, metadata=metadata_dict)

    return ps


def _process_psd_data(ps, metadata):
    """
    Create an xarray DataArray for a PSD dataset.

    Inputs:
        ps       = PSII_data instance
        metadata = INF file metadata dictionary

    :param ps: plantcv.plantcv.classes.PSII_data
    :param metadata: dict
    """
    bin_filepath = _dat_filepath(dataset="PSD", datapath=ps.datapath, filename=ps.filename)
    if os.path.exists(bin_filepath):
        img_cube, frame_labels, frame_nums = _read_dat_file(dataset="PSD", filename=bin_filepath,
                                                            height=int(metadata["ImageRows"]),
                                                            width=int(metadata["ImageCols"]))
        # If not all frames are saved the order is fixed
        # Phenovation does not update the framenumbers in the references.
        # Default frames (when SaveAllFrames == 0)
        f0_frame = 1
        fm_frame = 2
        # If the method is labeled FvFm
        if 'FvFmFrameF0' in metadata and metadata["SaveAllFrames"] != "0":
            f0_frame = int(metadata["FvFmFrameF0"]) + 1
            fm_frame = int(metadata["FvFmFrameFm"]) + 1
        # If the method is labeled DkOjip
        if 'DkOjipFrameF0' in metadata and metadata["SaveAllFrames"] != "0":
            f0_frame = int(metadata["DkOjipFrameF0"]) + 1  # data cube includes Fdark at the beginning
            fm_frame = int(metadata["DkOjipFrameFm"]) + 1
        frame_labels[0] = 'Fdark'
        frame_labels[f0_frame] = 'F0'
        frame_labels[fm_frame] = 'Fm'
        psd = xr.DataArray(
            data=img_cube[..., None],
            dims=('x', 'y', 'frame_label', 'measurement'),
            coords={'frame_label': frame_labels,
                    'frame_num': ('frame_label', frame_nums),
                    'measurement': ['t0']},
            name='ojip_dark'
        )
        psd.attrs["long_name"] = "OJIP dark-adapted measurements"
        ps.add_data(psd)

        _debug(visual=ps.ojip_dark.squeeze('measurement', drop=True),
               filename=os.path.join(params.debug_outdir, f"{str(params.device)}_PSD-frames.png"),
               col='frame_label',
               col_wrap=int(np.ceil(ps.ojip_dark.frame_label.size / 4)))


def _process_psl_data(ps, metadata):
    """
    Create an xarray DataArray for a PSL dataset.

    Inputs:
        ps       = PSII_data instance
        metadata = INF file metadata dictionary

    :param ps: plantcv.plantcv.classes.PSII_data
    :param metadata: dict
    """
    bin_filepath = _dat_filepath(dataset="PSL", datapath=ps.datapath, filename=ps.filename)
    if os.path.exists(bin_filepath):
        img_cube, frame_labels, frame_nums = _read_dat_file(dataset="PSL", filename=bin_filepath,
                                                            height=int(metadata["ImageRows"]),
                                                            width=int(metadata["ImageCols"]))
        # If not all frames are saved the order is fixed
        # Phenovation does not update the framenumbers in the references.
        # Default frames (when SaveAllFrames == 0)
        fsp_frame = 1
        fmp_frame = 2
        # If the method is labeled FqFm
        if 'FqFmFrameFsp' in metadata and metadata["SaveAllFrames"] != "0":
            fsp_frame = int(metadata["FqFmFrameFsp"]) + 1
            fmp_frame = int(metadata["FqFmFrameFmp"]) + 1
        # If the method is labeled LtOjip
        if 'LtOjipFrameFsp' in metadata and metadata["SaveAllFrames"] != "0":
            fsp_frame = int(metadata["LtOjipFrameFsp"]) + 1
            fmp_frame = int(metadata["LtOjipFrameFmp"]) + 1
        frame_labels[0] = "Flight"
        frame_labels[fsp_frame] = 'Fp'
        frame_labels[fmp_frame] = 'Fmp'
        psl = xr.DataArray(
            data=img_cube[..., None],
            dims=('x', 'y', 'frame_label', 'measurement'),
            coords={'frame_label': frame_labels,
                    'frame_num': ('frame_label', frame_nums),
                    'measurement': ['t1']},
            name='ojip_light'
        )
        psl.attrs["long_name"] = "OJIP light-adapted measurements"
        ps.add_data(psl)

        _debug(visual=ps.ojip_light.squeeze('measurement', drop=True),
               filename=os.path.join(params.debug_outdir, f"{str(params.device)}_PSL-frames.png"),
               col='frame_label',
               col_wrap=int(np.ceil(ps.ojip_light.frame_label.size / 4)))


def _process_pmd_data(ps, metadata):
    """
    Create an xarray DataArray for a PMD dataset.

    Inputs:
        ps       = PSII_data instance
        metadata = INF file metadata dictionary

    :param ps: plantcv.plantcv.classes.PSII_data
    :param metadata: dict
    """
    bin_filepath = _dat_filepath(dataset="PMD", datapath=ps.datapath, filename=ps.filename)
    if os.path.exists(bin_filepath):
        img_cube, frame_labels, frame_nums = _read_dat_file(dataset="PMD", filename=bin_filepath,
                                                            height=int(metadata["ImageRows"]),
                                                            width=int(metadata["ImageCols"]))
        frame_labels = ["Fdark", "F0", "Fm", "Fs"]
        pmd = xr.DataArray(
            data=img_cube[..., None],
            dims=('x', 'y', 'frame_label', 'measurement'),
            coords={'frame_label': frame_labels,
                    'frame_num': ('frame_label', frame_nums),
                    'measurement': ['t0']},
            name='pam_dark'
        )
        pmd.attrs["long_name"] = "pam dark-adapted measurements"
        ps.add_data(pmd)

        _debug(visual=ps.pam_dark.squeeze('measurement', drop=True),
               filename=os.path.join(params.debug_outdir, f"{str(params.device)}_PMD-frames.png"),
               col='frame_label',
               col_wrap=int(np.ceil(ps.pam_dark.frame_label.size / 4)))


def _process_pml_data(ps, metadata):
    """
    Create an xarray DataArray for a PML dataset.

    Inputs:
        ps       = PSII_data instance
        metadata = INF file metadata dictionary

    :param ps: plantcv.plantcv.classes.PSII_data
    :param metadata: dict
    """
    bin_filepath = _dat_filepath(dataset="PML", datapath=ps.datapath, filename=ps.filename)
    if os.path.exists(bin_filepath):
        img_cube, frame_labels, frame_nums = _read_dat_file(dataset="PML", filename=bin_filepath,
                                                            height=int(metadata["ImageRows"]),
                                                            width=int(metadata["ImageCols"]))
        frame_labels = ["Flight", "Fp", "Fmp", "Fs"]
        pml = xr.DataArray(
            data=img_cube[..., None],
            dims=('x', 'y', 'frame_label', 'measurement'),
            coords={'frame_label': frame_labels,
                    'frame_num': ('frame_label', frame_nums),
                    'measurement': ['t0']},
            name='pam_light'
        )
        pml.attrs["long_name"] = "pam light-adapted measurements"
        ps.add_data(pml)

        _debug(visual=ps.pam_light.squeeze('measurement', drop=True),
               filename=os.path.join(params.debug_outdir, f"{str(params.device)}_PML-frames.png"),
               col='frame_label',
               col_wrap=int(np.ceil(ps.pam_light.frame_label.size / 4)))


def _process_chl_data(ps, metadata):
    """
    Create an xarray DataArray for a CHL dataset.

    Inputs:
        ps       = PSII_data instance
        metadata = INF file metadata dictionary

    :param ps: plantcv.plantcv.classes.PSII_data
    :param metadata: dict
    """
    bin_filepath = _dat_filepath(dataset="CHL", datapath=ps.datapath, filename=ps.filename)
    if os.path.exists(bin_filepath):
        img_cube, frame_labels, _ = _read_dat_file(dataset="CHL", filename=bin_filepath,
                                                   height=int(metadata["ImageRows"]),
                                                   width=int(metadata["ImageCols"]))
        frame_labels = ["Fdark", "Chl"]
        chl = xr.DataArray(
            data=img_cube,
            dims=('x', 'y', 'frame_label'),
            coords={'frame_label': frame_labels},
            name='chlorophyll'
        )
        chl.attrs["long_name"] = "chlorophyll measurements"
        ps.add_data(chl)

        _debug(visual=ps.chlorophyll,
               filename=os.path.join(params.debug_outdir, f"{str(params.device)}_CHL-frames.png"),
               col='frame_label',
               col_wrap=int(np.ceil(ps.chlorophyll.frame_label.size / 4)))


def _process_spc_data(ps, metadata):
    """
    Create a Spectral_data object for the SPC and CLR datasets.

    Inputs:
        ps       = PSII_data instance
        metadata = INF file metadata dictionary

    :param ps: plantcv.plantcv.classes.PSII_data
    :param metadata: dict
    """
    img_cubes = []
    wavelengths = []
    clr_filepath = _dat_filepath(dataset="CLR", datapath=ps.datapath, filename=ps.filename)
    spc_filepath = _dat_filepath(dataset="SPC", datapath=ps.datapath, filename=ps.filename)
    rgb = None
    if os.path.exists(clr_filepath):
        rgb_cube, _, _ = _read_dat_file(dataset="CLR", filename=clr_filepath,
                                        height=int(metadata["ImageRows"]),
                                        width=int(metadata["ImageCols"]))
        img_cubes.append(rgb_cube)
        wavelengths += [640, 550, 475]
        rgb = img_as_ubyte(rgb_cube[:, :, [2, 1, 0]])
    if os.path.exists(spc_filepath):
        spc_cube, _, _ = _read_dat_file(dataset="SPC", filename=spc_filepath,
                                        height=int(metadata["ImageRows"]),
                                        width=int(metadata["ImageCols"]))
        img_cubes.append(spc_cube)
        wavelengths += [540, 710, 770]
        if rgb is None:
            rgb = img_as_ubyte(spc_cube)

    if len(img_cubes) > 0:
        if len(img_cubes) == 2:
            # Concatenate the images on the depth/spectral (z) axis
            array_data = np.concatenate(img_cubes, axis=2)
        else:
            array_data = img_cubes[0]

        # sort all wavelengths
        wavelengths = np.array(wavelengths)
        ind = np.argsort(wavelengths)
        wavelengths = wavelengths[ind]

        wavelength_dict = {}
        for (idx, wv) in enumerate(wavelengths):
            wavelength_dict[wv] = float(idx)

        # sort array_data based on wavelengths
        array_data = array_data[:, :, ind]
        # Scale the array data to 0-1 by dividing by the maximum data type value
        array_data = (array_data / np.iinfo(array_data.dtype).max).astype(np.float32)

        # Create a Spectral_data object
        rows, columns = array_data.shape[0:2]
        multispec = Spectral_data(array_data=array_data,
                                  max_wavelength=float(max(wavelengths)),
                                  min_wavelength=float(min(wavelengths)),
                                  max_value=float(np.amax(array_data)),
                                  min_value=float(np.amin(array_data)),
                                  d_type=array_data.dtype,
                                  wavelength_dict=wavelength_dict,
                                  samples=columns, lines=rows, interleave="NA",
                                  wavelength_units="nm", array_type="multispectral",
                                  pseudo_rgb=rgb, filename="NA", default_bands=None)

        ps.spectral = multispec

        _debug(visual=ps.spectral.pseudo_rgb,
               filename=os.path.join(params.debug_outdir, f"{str(params.device)}_spectral-RGB.png"))


def _dat_filepath(dataset, datapath, filename):
    """
    Create the filepath to a DAT file based on the INF filename.

    Inputs:
        dataset  = dataset key (PSD, PSL, SPC, CHL, CLR)
        datapath = path to the dataset (basepath of the INF file)
        filename = INF filename

    Returns:
        bin_filepath = fully-qualified path to the DAT file

    :param dataset: str
    :param datapath: str
    :param filename: str
    :return bin_filepath: str
    """
    filename_components = filename.split("_")
    # Find corresponding bin img filepath based on .INF filepath
    # replace header with bin img type
    filename_components[filename_components.index('HDR')] = dataset
    bin_filenames = "_".join(filename_components)
    bin_filename = bin_filenames.replace(".INF", ".DAT")
    bin_filepath = os.path.join(datapath, bin_filename)

    return bin_filepath


def _read_dat_file(dataset, filename, height, width):
    """
    Read raw data from DAT file.

    Inputs:
        dataset  = dataset key (PSD, PSL, SPC, CHL, CLR)
        filename = fully-qualified path to the DAT file
        height   = height (rows) of the images
        width    = width (columns) of the image

    Returns:
        img_cube     = raw data cube in NumPy shape
        frame_labels = list of labels for each frame
        frame_nums   = the number of frames

    :param dataset: str
    :param filename: str
    :param height: int
    :param width: int
    :return img_cube: numpy.ndarray
    :return frame_labels: list
    :return frame_numbs: int
    """
    print(f'Compiling: {dataset}')
    # Dump in bin img data
    raw_data = np.fromfile(filename, np.uint16, -1)
    # Reshape, numpy shaped
    img_cube = raw_data.reshape(int(len(raw_data) / (height * width)), width, height).transpose((2, 1, 0))

    # Calculate frames of interest and keep track of their labels. labels must be unique across all measurements
    frame_labels = [(dataset + str(i)) for i in range(img_cube.shape[2])]
    frame_nums = np.arange(img_cube.shape[2])

    return img_cube, frame_labels, frame_nums



# Perform quality control by checking for problematic color data
import os
import cv2
import numpy as np
import pandas as pd
import altair as alt
from plantcv.plantcv import outputs, params, warn
from plantcv.plantcv._debug import _debug


# Function to check for over- or underexposure
def _check_exposure(channel, warning_threshold, channel_name):
    """Check if a color channel is over- or underexposed.

    This function analyzes the given color channel to determine if
    more than the specified percentage of its pixels are either at the
    minimum (0) or maximum (255) intensity values, which may indicate
    over- or underexposure issues.

    Parameters
    ----------
    channel : numpy.ndarray
        A 2D numpy array representing the color channel of an image.
    warning_threshold : float
        The threshold value for triggering a warning for over- or underexposure.
    channel_name : str
        Name of the channel being analyzed (e.g., "red", "green", "blue").

    Returns
    -------
    bool
        True if the channel is over- or underexposed; False otherwise.
    """
    total_pixels = channel.size
    zero_count = np.sum(channel == 0)
    max_count = np.sum(channel == 255)
    proportion_bad_pix = zero_count / total_pixels
    outputs.add_metadata(term=f"{channel_name}_percent_bad_exposure_qc", datatype=float, value=proportion_bad_pix)
    return (zero_count / total_pixels > warning_threshold) or (max_count / total_pixels > warning_threshold)


def exposure(rgb_img, warning_threshold=0.05):
    """Perform quality control by checking for problematic color data and plotting histograms.

    This function performs an analysis of an image to check for over- or underexposure
    in the red, green, and blue color channels. It also generates and displays histograms
    for each color channel to visualize the distribution of pixel intensities.

    Parameters
    ----------
    rgb_img : numpy.ndarray
        Color image data.
    warning_threshold : float, optional
        The threshold value for triggering a warning for over- or underexposure, by default 0.05

    Returns:
    -------
    chart : altair.vegalite.v5.api.Chart
        Histogram chart of RGB image intensity values.
    """
    params.device += 1
    # Convert the img from BGR to RGB
    img_rgb = cv2.cvtColor(rgb_img, cv2.COLOR_BGR2RGB)

    # Split the img into its Red, Green, and Blue channels
    red_channel, green_channel, blue_channel = img_rgb[:, :, 0], img_rgb[:, :, 1], img_rgb[:, :, 2]

    # Check each channel for over- or underexposure
    if (
        _check_exposure(red_channel, warning_threshold, channel_name="red") or
        _check_exposure(green_channel, warning_threshold, channel_name="green") or
        _check_exposure(blue_channel, warning_threshold, channel_name="blue")
    ):
        warn(
            f"The image is over- or underexposed because more than {warning_threshold * 100}% of "
            "pixels are equal to 0 or 255 intensity. Color cannot be analyzed "
            "responsibly, as color values are lost above the minimum (0) and maximum "
            "(255). Change camera settings to capture appropriate images.")

    # Create a dataframe to store the histogram data
    df = pd.DataFrame({
        'intensity': np.arange(256),
        'Red Channel': np.histogram(red_channel.ravel(), bins=256)[0] / red_channel.size,
        'Green Channel': np.histogram(green_channel.ravel(), bins=256)[0] / green_channel.size,
        'Blue Channel': np.histogram(blue_channel.ravel(), bins=256)[0] / blue_channel.size
    })

    # Plot the histograms
    chart = alt.Chart(df).transform_fold(
        ["Red Channel", "Green Channel", "Blue Channel"],
        as_=['Channel', 'Proportion']
    ).mark_area(
        opacity=0.5
    ).encode(
        alt.X('intensity:Q', title="Intensity Value"),
        alt.Y('Proportion:Q', title="Proportion of Pixels"),
        alt.Color(
            'Channel:N',
            scale=alt.Scale(range=['red', 'green', 'blue']),
            sort=['red', 'green', 'blue'],
            legend=None
        ),
        column=alt.Column('Channel:N', sort=['Red Channel', 'Green Channel', 'Blue Channel'], title=None)
    ).properties(
        width=200,
        height=200
    )

    # Display or save the plot
    _debug(chart, filename=os.path.join(params.debug_outdir, f"{params.device}_bad_exposure_hist.png"))

    return chart



# Image Series Segmentation

import os
import math
import numpy as np
import cv2 as cv
from scipy import ndimage as ndi
from skimage.segmentation import watershed
from skimage.color import label2rgb
from plantcv.plantcv import readimage
from plantcv.plantcv import fill_holes
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _rgb2gray


def segment_image_series(imgs_paths, masks_paths, rois, save_labels=True, ksize=3):
    """Segments the objects in a time series of images using watershed segmentation.
    The objects (labels) are given by a list of rois (region of interest) and the labels
    are propagated sequentially in the time dimension using blocks of ksize.

    Inputs:
    imgs_paths  = List of paths to the images in the time series. Ordered by time
    masks_paths = List of paths to the masks in the time series.
                  Each mask should correspond to the image in imgs_paths for the same index
    rois        = List of roi contours
    save_labels = Optional, saves the labels of each image independently
    ksize       = Size of the block in the time dimension to propagate the labels

    Returns:
    out_labels = 3D array containing the labels of the whole time series

    :param imgs_paths:  list
    :param masks_paths: list
    :param rois:        list
    :param save_labels: bool
    :param ksize:       int
    :return out_labels: numpy.ndarray
    """
    debug = params.debug
    params.debug = None
    params.color_sequence = 'random'

    # for symmetry, using blocks (kernels) of size 2*floor(ksize/2) + 1
    half_k = math.floor(ksize/2)

    image_names = [os.path.basename(img_path) for img_path in imgs_paths]

    # get the size of the images
    tmp, _, _ = readimage(filename=masks_paths[0])
    h, w = tmp.shape[0], tmp.shape[1]

    # create an image where all the pixels inside each roi have the roi label
    roi_labels = np.zeros((h, w), dtype=np.uint8)
    n_labels = len(rois)
    for i in range(n_labels):
        img_roi = np.zeros((h, w), dtype=np.uint8)
        img_roi = cv.drawContours(img_roi, rois[i], -1, 255, 3)
        img_roi = fill_holes(img_roi)
        roi_labels = roi_labels + (img_roi == 255)*(i+1)

    # output initialization
    N = len(image_names)
    out_labels = np.zeros((h, w, N), dtype=np.uint8)
    out_labels[:, :, 0] = roi_labels.copy()

    # Propagate labels sequentially n is the index in the output of the frame currently
    # in process. At each iteration only one frame is labeled.
    for n in range(0, N):
        # size of the stack used at each iteration
        d = 2*half_k+1

        # stacks init
        img_stack = np.zeros((h, w, d))
        mask_stack = np.zeros((h, w, d))
        markers = np.zeros((h, w, d), dtype=np.int32)

        # The number of frames used is always the same but the borders are
        # treated as 'constant' or 'zero padding'
        stack_idx = 0  # borders are 'constant'

        # loop to build the stacks. half_k gives the index of the frame in process
        for m in range(-half_k, half_k+1):
            frame = min(N-1, max(n+m, 0))  # border handling

            img, _, _ = readimage(filename=imgs_paths[frame])
            img_stack[:, :, stack_idx] = _rgb2gray(rgb_img=img)
            mask, _, _ = readimage(filename=masks_paths[frame], mode='gray')
            mask_stack[:, :, stack_idx] = mask

            if m == 0:
                # required to create the output image
                img_n_rgb = img
                # enforcing the label inside the regions of interest
                markers[:, :, stack_idx] = (mask != 0)*roi_labels
            else:
                markers[:, :, stack_idx] = out_labels[:, :, frame]

            stack_idx += 1

        # edges using 3D sobel operator as elevation map for watershed
        edges = ndi.generic_gradient_magnitude(img_stack, ndi.sobel)

        # segmentation using the watershed algorithm
        labels = watershed(edges, markers=markers, mask=mask_stack, compactness=0)

        # add the resulting labels to the outputs block
        out_labels[:, :, n] = labels[:, :, half_k]

        # create images for plotting and printing (debug mode)
        vis_seg = label2rgb(out_labels[:, :, n], image=img_n_rgb, colors=None, alpha=0.3, bg_label=0)
        # cast visualization image as int
        vis_seg = (np.floor(255*vis_seg)).astype(np.uint8)
        if n == N - 1:
            params.debug = debug
            _debug(visual=vis_seg, filename=os.path.join(params.debug_outdir,
                                                         f"{str(params.device)}_{image_names[n][:-4]}_WSeg.png"))

    if save_labels is True:
        for i in range(N):
            np.save(os.path.join(params.debug_outdir, f"{image_names[i][:-4]}_labels"), out_labels[:, :, i])

    return out_labels



import os
import sys
import json
import datetime


class WorkflowConfig:
    """PlantCV Parallel Configuration class"""

    def __init__(self):
        self.input_dir = ""
        self.json = ""
        self.filename_metadata = []
        self.workflow = ""
        self.img_outdir = "./output_images"
        self.include_all_subdirs = True
        self.tmp_dir = "."
        self.start_date = None
        self.end_date = None
        self.imgformat = "all"
        self.delimiter = "_"
        self.metadata_filters = {}
        self.metadata_regex = {}
        self.timestampformat = "%Y-%m-%dT%H:%M:%S.%fZ"
        self.writeimg = False
        self.other_args = {}
        self.groupby = ["filepath"]
        self.group_name = "auto"
        self.cleanup = True
        self.append = False
        self.cluster = "LocalCluster"
        self.cluster_config = {
            "n_workers": 1,
            "cores": 1,
            "memory": "1GB",
            "disk": "1GB",
            "log_directory": None,
            "local_directory": None,
            "job_extra_directives": None
        }
        self.metadata_terms = self.metadata_term_definition()
    # set metadata_terms reactively based on filename_metadata

    @property
    def metadata_terms(self):
        self._metadata_terms = self.metadata_term_definition()
        return self._metadata_terms

    @metadata_terms.setter
    def metadata_terms(self, new):
        self._metadata_terms = new

    # Save configuration to a file
    def save_config(self, config_file):
        """Save configuration to a file.

        Input variables:
        config_file = Filename to write configuration to (text/JSON)

        :param config_file: str
        """
        # Open the file for writing
        with open(config_file, "w") as fp:
            # strip the underscore from _metadata_terms before writing
            content = vars(self)
            content = {k.strip("_"): v for k, v in content.items()}
            # Save the data in JSON format with indentation
            json.dump(obj=content, fp=fp, indent=4)

    # Import a configuration from a file
    def import_config(self, config_file):
        """Import a configuration file.

        Input variables:
        config_file = Configuration file to import

        :param config_file: str
        """
        # Open the file for reading
        with open(config_file, "r") as fp:
            # Import the JSON configuration data
            config = json.load(fp)
            for key, value in config.items():
                if key != "_metadata_terms":
                    setattr(self, key, value)

    # Validation checks on current config
    def validate_config(self):
        """Validation checks on current configuration."""
        checks = [True]
        # Validate input directory
        if not os.path.exists(self.input_dir):
            print(f"Error: input directory (input_dir) is required and {self.input_dir} does not exist.",
                  file=sys.stderr)
            checks.append(False)
        # Validate JSON file
        if self.json == "":
            print("Error: an output JSON file (json) is required but is currently undefined.", file=sys.stderr)
            checks.append(False)
        # Validate workflow script
        if not os.path.exists(self.workflow):
            print(f"Error: PlantCV workflow script (workflow) is required and {self.workflow} does not exist.",
                  file=sys.stderr)
            checks.append(False)

        # Validate start_date and end_date formats
        if self.start_date is not None:
            try:
                _ = datetime.datetime.strptime(
                    self.start_date, self.timestampformat)
            except ValueError as e:
                print(str(e) + '\n  --> Please specify the start_date according to the timestampformat  <--\n')
                checks.append(False)
        if self.end_date is not None:
            try:
                _ = datetime.datetime.strptime(
                    self.end_date, self.timestampformat)
            except ValueError as e:
                print(str(
                    e) + '\n  --> Please specify the end_date according to the timestampformat  <--\n')
                checks.append(False)

        # Validate the cluster type
        valid_clusters = ["HTCondorCluster", "LocalCluster", "LSFCluster", "MoabCluster", "OARCluster", "PBSCluster",
                          "SGECluster", "SLURMCluster"]
        if self.cluster not in valid_clusters:
            print(f"Error: the cluster type {self.cluster} is not a supported cluster provider. "
                  f"Valid clusters include: {', '.join(map(str, valid_clusters))}."
                  )
        return all(checks)

    # Specify metadata terms
    def metadata_term_definition(self):
        """Add dictionary of metadata terms"""
        metadata_terms = {
            "timestamp": {
                "label": "datetime of image",
                "datatype": "<class 'datetime.datetime'>",
                "value": None
            }
        }
        default_metadata_terms = {
            # Camera settings
            "camera": {
                "label": "camera identifier",
                "datatype": "<class 'str'>",
                "value": "none"
            },
            "imgtype": {
                "label": "image type",
                "datatype": "<class 'str'>",
                "value": "none"
            },
            "zoom": {
                "label": "camera zoom setting",
                "datatype": "<class 'str'>",
                "value": "none"
            },
            "exposure": {
                "label": "camera exposure setting",
                "datatype": "<class 'str'>",
                "value": "none"
            },
            "gain": {
                "label": "camera gain setting",
                "datatype": "<class 'str'>",
                "value": "none"
            },
            "frame": {
                "label": "image series frame identifier",
                "datatype": "<class 'str'>",
                "value": "none"
            },
            "rotation": {
                "label": "sample rotation in degrees",
                "datatype": "<class 'str'>",
                "value": "none"
            },
            "lifter": {
                "label": "imaging platform height setting",
                "datatype": "<class 'str'>",
                "value": "none"
            },
            # Date-Time
            "timestamp": {
                "label": "datetime of image",
                "datatype": "<class 'datetime.datetime'>",
                "value": None
            },
            # Sample attributes
            "id": {
                "label": "image identifier",
                "datatype": "<class 'str'>",
                "value": "none"
            },
            "barcode": {
                "label": "plant barcode identifier",
                "datatype": "<class 'str'>",
                "value": "none"
            },
            "treatment": {
                "label": "treatment identifier",
                "datatype": "<class 'str'>",
                "value": "none"
            },
            "cartag": {
                "label": "plant carrier identifier",
                "datatype": "<class 'str'>",
                "value": "none"
            },
            # Experiment attributes
            "measurementlabel": {
                "label": "experiment identifier",
                "datatype": "<class 'str'>",
                "value": "none"
            },
            # Other
            "other": {
                "label": "other identifier",
                "datatype": "<class 'str'>",
                "value": "none"
            }
        }
        # add any other metadata terms as strings
        for k in [fm for fm in self.filename_metadata if fm not in default_metadata_terms]:
            metadata_terms[k] = {
                "label": f"{k}",
                "datatype": "<class 'str'>",
                "value": "none"
            }
        # add default terms with their formatting
        for k in [fm for fm in self.filename_metadata if fm in default_metadata_terms]:
            metadata_terms[k] = default_metadata_terms[k]

        return metadata_terms



"""Find geodesic lengths of skeleton segments."""
import os
import cv2
from plantcv.plantcv import params, outputs
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _scale_size


def segment_path_length(segmented_img, objects, label=None):
    """Use segments to calculate geodesic distance per segment.

    Inputs:
    segmented_img = Segmented image to plot lengths on
    objects       = List of contours
    label         = Optional label parameter, modifies the variable name of
                    observations recorded (default = pcv.params.sample_label).

    Returns:
    labeled_img        = Segmented debugging image with lengths labeled

    :param segmented_img: numpy.ndarray
    :param objects: list
    :param label: str
    :return labeled_img: numpy.ndarray
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    label_coord_x = []
    label_coord_y = []
    segment_lengths = []
    labeled_img = segmented_img.copy()

    for obj in objects:
        # Calculate geodesic distance, divide by two since cv2 seems to be taking the perimeter of the contour
        segment_lengths.append(float(cv2.arcLength(obj, False) / 2))
        # Store coordinates for labels
        label_coord_x.append(obj[0][0][0])
        label_coord_y.append(obj[0][0][1])

    segment_ids = []
    # Put labels of length
    for c, value in enumerate(segment_lengths):
        text = f"{value:0,.2f}"
        w = label_coord_x[c]
        h = label_coord_y[c]
        cv2.putText(img=labeled_img, text=text, org=(w, h), fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                    fontScale=params.text_size, color=(150, 150, 150), thickness=params.text_thickness)
        segment_ids.append(c)

    outputs.add_observation(sample=label, variable='segment_path_length', trait='segment path length',
                            method='plantcv.plantcv.morphology.segment_path_length', scale=params.unit, datatype=list,
                            value=_scale_size(segment_lengths), label=segment_ids)

    _debug(visual=labeled_img, filename=os.path.join(params.debug_outdir, f"{params.device}_segment_path_lengths.png"))

    return labeled_img



# Detect dics based on object eccentricity
import os
import numpy as np
from skimage.measure import label, regionprops
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug


def eccentricity(bin_img, ecc_thresh=0):
    """Detect/filter disc-shaped regions in a binary image based on eccentricity.

    A value of eccentricity between 0 and 1 corresponds to an ellipse.
    The closer the value to 0 the closer the shape is to a circle.

    Inputs:
    bin_img       = Binary image containing the connected regions to consider
    ecc_thresh    = Eccentricity threshold below which a region is kept


    Returns:
    discs_mask  = Binary image that contains only the detected discs

    :param bin_img: numpy.ndarray
    :param ecc_thresh: float
    :return discs_mask: numpy.ndarray
    """
    params.device += 1
    # label connected regions
    labeled_img = label(bin_img)
    # measure region properties
    obj_measures = regionprops(labeled_img)
    # blank mask to draw discs onto
    discs_mask = np.zeros(labeled_img.shape, dtype=np.uint8)
    # Store the list of coordinates (row,col) for the objects that pass
    for obj in obj_measures:
        if obj.eccentricity < ecc_thresh:
            # Convert coord values to int
            discs_mask += np.where(labeled_img == obj.label, 255, 0).astype(np.uint8)

    _debug(visual=discs_mask, filename=os.path.join(params.debug_outdir, f"{params.device}_discs_mask_{ecc_thresh*10}.png"))

    return discs_mask



# Calculate masked average background reflectance

import numpy as np


def _avg_reflectance(spectral_data, mask):
    """Find average reflectance of masked hyperspectral data instance.
    This is useful for calculating a target signature (n_band x 1 - column array) which is required in various GatorSense
    hyperspectral tools (https://github.com/GatorSense/hsi_toolkit_py)

    Inputs:
    spectral_array = Hyperspectral data instance
    mask           = Target wavelength value

    Returns:
    idx            = Index

    :param spectral_data: __main__.Spectral_data
    :param mask: numpy.ndarray
    :return spectral_array: __main__.Spectral_data
    """
    # Initialize list of average reflectance values
    avg_r = []

    # For each band in a hyperspectral datacube mask and take the average
    for i in range(0, len(spectral_data.wavelength_dict)):
        band = spectral_data.array_data[:, :, [i]]
        band_reshape = np.transpose(np.transpose(band)[0])
        masked_band = band_reshape[np.where(mask > 0)]
        band_avg = np.average(masked_band)
        avg_r.append(band_avg)

    # Convert into array object rather than list
    avg_r = np.asarray(avg_r)

    return avg_r



"""Analyze the vertical distribution of the plant relative to a horizontal reference line."""
import os
import cv2
import numpy as np
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _iterate_analysis, _grayscale_to_rgb, _scale_size
from plantcv.plantcv import params
from plantcv.plantcv import outputs


def _get_boundary_values(bound_mask, total_area, axis=0):
    """
    Calculate area, percent area, and distance (height or width) in a boundary.

    Parameters
    ----------
    bound_mask : numpy.ndarray
        A binary mask of one side of a horizontal or vertical boundary.
    total_area : int
        Total area of the unbounded mask.
    axis : int, optional
        Which axis to use in calculated distance. 0 is height, 1 is width.

    Returns
    -------
    bound_area : int
        Area within boundary.
    distance_bound : int
        Height/width of mask within boundary.
    percent_area_bound : float
        Percent of total area in this boundary.
    """
    bound_area = int(np.sum(bound_mask))
    distance_bound = 0
    percent_area_bound = 0
    if bound_area:
        distance_bound = int(np.max(np.where(bound_mask)[axis]) +
                             1 - np.min(np.where(bound_mask)[axis]))
        percent_area_bound = int((bound_area / total_area) * 100)
    return bound_area, distance_bound, percent_area_bound


def _boundary_img_annotation(img, mask, line_position, axis=0):
    """Annotate a debug image used in horizontal/vertical boundary analysis
    Parameters
    ----------
    img : numpy.ndarray
        RGB or grayscale image data for plotting.
    mask : numpy.ndarray
        Binary mask of objects (32-bit).
    line_position : int
        Position of boundary line in pixels from top to bottom (a value of 0 draws the line through the top of the image).
    axis : int
        Which axis to use in drawing division

    Returns
    -------
    out_image : numpy.ndarray
        Diagnostic image showing measurements.
    """
    out_img = np.copy(img)
    # split mask by boundary
    mask1 = np.copy(mask).astype(bool)
    mask2 = np.copy(mask).astype(bool)
    # split mask by line_position on axis
    if not bool(axis):
        # fill in area on opposite side of threshold with 0s
        mask1[line_position:np.shape(mask1)[0] + 1] = np.zeros(
            np.shape(mask1[line_position:np.shape(mask1)[0] + 1]))
        mask2[0:line_position - 1] = np.zeros(
            np.shape(mask2[0:line_position - 1]))
    else:
        # fill in area on opposite side of threshold
        mask1[:, line_position:np.shape(mask1)[1] + 1] = np.zeros(
            np.shape(mask1[:, line_position:np.shape(mask1)[1] + 1]))
        mask2[:, 0:line_position - 1] = np.zeros(
            np.shape(mask2[:, 0:line_position - 1]))

    # replace mask with colors
    out_img[np.where(mask1)] = (255, 0, 255)
    out_img[np.where(mask2)] = (0, 255, 0)
    # draw boundary line
    line_start = [(0, line_position), (line_position, 0)][axis]
    line_end = [(np.shape(out_img)[1], line_position), (line_position, np.shape(img)[0])][axis]
    cv2.line(out_img, line_start, line_end, (255, 0, 0), thickness=params.line_thickness)
    return out_img


def bound_horizontal(img, labeled_mask, line_position, n_labels=1, label=None):
    """
    Analyze the vertical distribution of the plant relative to a horizontal reference line.

    Parameters
    ----------
    img : numpy.ndarray
        RGB or grayscale image data for plotting.
    labeled_mask : numpy.ndarray
        Labeled mask of objects (32-bit).
    n_labels : int, optional
        Total number of expected individual objects (default = 1).
    line_position : int
        Position of boundary line in pixels from top to bottom (a value of 0 draws the line through the top of the image).
    label : str, optional
        Optional label parameter, modifies the variable name of observations recorded (default = pcv.params.sample_label).

    Returns
    -------
    analysis_image : numpy.ndarray
        Diagnostic image showing measurements.
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    img = _iterate_analysis(img=img, labeled_mask=labeled_mask, n_labels=n_labels,
                            label=label, function=_analyze_bound_horizontal,
                            **{"line_position": line_position})
    img = _boundary_img_annotation(img, labeled_mask, line_position, 0)
    # Debugging
    _debug(visual=img, filename=os.path.join(params.debug_outdir, str(params.device) + '_boundary_on_img.png'))
    return img


def _analyze_bound_horizontal(img, mask, line_position, label):
    """
    User-input boundary line analysis for individual objects.

    Parameters
    ----------
    img : numpy.ndarray
        RGB or grayscale image data for plotting.
    mask : numpy.ndarray
        Binary image data.
    line_position : int
        Position of boundary line in pixels from top to bottom.
    label : str
        Label of object.

    Returns
    -------
    ori_img : numpy.ndarray
        Output image.
    """
    # Initialize output measurements
    height_above_bound = 0
    height_below_bound = 0
    above_bound_area = 0
    percent_bound_area_above = 0
    below_bound_area = 0
    percent_bound_area_below = 0
    ori_img = np.copy(img)
    # Skip empty masks
    if np.count_nonzero(mask) != 0 and line_position >= 0:
        ori_img = _grayscale_to_rgb(ori_img)
        # make copy of mask for above and below threshold
        top_mask = np.copy(mask).astype(bool)
        bottom_mask = np.copy(mask).astype(bool)
        # fill in area on opposite side of threshold with 0s
        top_mask[line_position:np.shape(top_mask)[0] + 1] = np.zeros(
            np.shape(top_mask[line_position:np.shape(top_mask)[0] + 1]))
        bottom_mask[0:line_position - 1] = np.zeros(
            np.shape(top_mask[0:line_position - 1]))
        tot_area = np.sum(mask.astype(bool))
        # calculate values above and below boundary
        above_bound_area, height_above_bound, percent_bound_area_above = _get_boundary_values(top_mask, tot_area, 0)
        below_bound_area, height_below_bound, percent_bound_area_below = _get_boundary_values(bottom_mask, tot_area, 0)

    outputs.add_observation(sample=label, variable='horizontal_reference_position',
                            trait='horizontal reference position',
                            method='plantcv.plantcv.analyze.bound_horizontal', scale='none', datatype=int,
                            value=line_position, label='none')
    outputs.add_observation(sample=label, variable='height_above_reference', trait='height above reference',
                            method='plantcv.plantcv.analyze.bound_horizontal', scale=params.unit, datatype=int,
                            value=_scale_size(height_above_bound), label=params.unit)
    outputs.add_observation(sample=label, variable='height_below_reference', trait='height_below_reference',
                            method='plantcv.plantcv.analyze.bound_horizontal', scale=params.unit, datatype=int,
                            value=_scale_size(height_below_bound), label=params.unit)
    outputs.add_observation(sample=label, variable='area_above_reference', trait='area above reference',
                            method='plantcv.plantcv.analyze.bound_horizontal', scale=params.unit, datatype=int,
                            value=_scale_size(above_bound_area, "area"), label=params.unit)
    outputs.add_observation(sample=label, variable='percent_area_above_reference', trait='percent area above reference',
                            method='plantcv.plantcv.analyze.bound_horizontal', scale='none', datatype=float,
                            value=percent_bound_area_above, label='none')
    outputs.add_observation(sample=label, variable='area_below_reference', trait='area below reference',
                            method='plantcv.plantcv.analyze.bound_horizontal', scale=params.unit, datatype=int,
                            value=_scale_size(below_bound_area, "area"), label=params.unit)
    outputs.add_observation(sample=label, variable='percent_area_below_reference', trait='percent area below reference',
                            method='plantcv.plantcv.analyze.bound_horizontal', scale='none', datatype=float,
                            value=percent_bound_area_below, label='none')

    return ori_img
